from __future__ import annotations


class HttpMongoError(RuntimeError):
    """Raised when the HTTP Mongo gateway returns an error."""


class HttpMongoTransportError(HttpMongoError):
    """Raised when the HTTP transport fails or returns invalid data."""



from __future__ import annotations

from collections import OrderedDict
from collections.abc import Iterable, Iterator, MutableMapping
from dataclasses import dataclass, field as dc_field
from typing import TYPE_CHECKING, Any, AnyStr, Mapping

import pyarrow as pa

from .field import Field, _normalize_metadata, _to_bytes

if TYPE_CHECKING:
    import polars as pl
    import pyspark.sql.types as pst


__all__ = [
    "Schema",
    "schema",
]


def schema(
    fields: Iterable[Field | pa.Field],
    *,
    metadata: dict[bytes | str, bytes | str | object] | None = None,
    tags: dict[bytes | str, bytes | str | object] | None = None,
) -> "Schema":
    return Schema.from_fields(
        fields,
        metadata=_normalize_metadata(metadata, tags=tags),
    )


@dataclass
class Schema(MutableMapping[str, Field]):
    inner_fields: OrderedDict[str, Field] = dc_field(default_factory=OrderedDict)
    metadata: dict[bytes, bytes] | None = None

    def __post_init__(self) -> None:
        if not isinstance(self.inner_fields, OrderedDict):
            normalized_fields: OrderedDict[str, Field] = OrderedDict()

            if isinstance(self.inner_fields, Mapping):
                for key, value in self.inner_fields.items():
                    field = Field.from_any(value)
                    if field.name != key:
                        field = field.copy(name=key)
                    normalized_fields[key] = field
            else:
                for value in self.inner_fields:
                    field = Field.from_any(value)
                    normalized_fields[field.name] = field

            self.inner_fields = normalized_fields

    @property
    def tags(self):
        if not self.metadata:
            return None

        return {k[2:]: v for k, v in self.metadata.items() if k.startswith(b"t:")}

    @tags.setter
    def tags(self, value: dict[AnyStr, AnyStr] | None):
        if value:
            if self.metadata is None:
                self.metadata = {}
            self.metadata.update(
                {
                    b"t:" + _to_bytes(k): _to_bytes(v)
                    for k, v in value.items()
                    if k and v
                }
            )

    def update_tags(self, value: dict[AnyStr, AnyStr] | None):
        if value:
            if self.metadata is None:
                self.metadata = {}
            self.metadata.update(
                {
                    b"t:" + _to_bytes(k): _to_bytes(v)
                    for k, v in value.items()
                    if k and v
                }
            )

    @property
    def names(self) -> tuple[str, ...]:
        return tuple(self.inner_fields.keys())

    @property
    def fields(self) -> tuple[Field, ...]:
        return tuple(self.inner_fields.values())

    @property
    def arrow_fields(self):
        return [_.to_arrow_field() for _ in self.inner_fields.values()]

    @property
    def partition_by(self):
        return [_ for _ in self.inner_fields.values() if _.partition_by]

    @property
    def cluster_by(self):
        return [_ for _ in self.inner_fields.values() if _.cluster_by]

    @property
    def comment(self):
        if not self.metadata:
            return None

        c = self.metadata.get(b"comment") or self.metadata.get(b"description")

        if not c:
            return None

        return c.decode("utf-8")

    @staticmethod
    def _merge_metadata(
        left: dict[bytes, bytes] | None,
        right: dict[bytes, bytes] | None,
    ) -> dict[bytes, bytes] | None:
        if not left and not right:
            return None
        return {**(left or {}), **(right or {})} or None

    @staticmethod
    def _coerce_other(other: Any) -> "Schema":
        if isinstance(other, Schema):
            return other
        return Schema.from_any(other)

    def copy(
        self,
        *,
        fields: Iterable[Field | pa.Field] | None = None,
        metadata: dict[bytes | str, bytes | str | object] | None = None,
        tags: dict[bytes | str, bytes | str | object] | None = None,
    ) -> "Schema":
        return Schema(
            inner_fields=(
                OrderedDict((field.name, Field.from_any(field)) for field in fields)
                if fields is not None
                else OrderedDict(
                    (name, field.copy()) for name, field in self.inner_fields.items()
                )
            ),
            metadata=(
                _normalize_metadata(metadata, tags=tags)
                if (metadata is not None or tags is not None)
                else (dict(self.metadata) if self.metadata else None)
            ),
        )

    def append(self, *more_fields: Field | pa.Field) -> "Schema":
        out = self.copy()
        for value in more_fields:
            field = Field.from_any(value)
            out[field.name] = field
        return out

    def extend(self, fields: Iterable[Field | pa.Field]) -> "Schema":
        out = self.copy()
        for value in fields:
            field = Field.from_any(value)
            out[field.name] = field
        return out

    def __getitem__(self, key: str) -> Field:
        return self.inner_fields[key]

    def __setitem__(self, key: str, value: Field | pa.Field) -> None:
        field = Field.from_any(value)
        if field.name != key:
            field = field.copy(name=key)
        self.inner_fields[key] = field

    def __delitem__(self, key: str) -> None:
        del self.inner_fields[key]

    def __iter__(self) -> Iterator[str]:
        return iter(self.inner_fields)

    def __len__(self) -> int:
        return len(self.inner_fields)

    def __contains__(self, key: object) -> bool:
        return key in self.inner_fields

    def __add__(self, other: Any) -> "Schema":
        other = self._coerce_other(other)

        merged_fields = OrderedDict(
            (name, field.copy()) for name, field in self.inner_fields.items()
        )
        for name, field in other.inner_fields.items():
            merged_fields[name] = field.copy()

        return Schema(
            inner_fields=merged_fields,
            metadata=self._merge_metadata(self.metadata, other.metadata),
        )

    def __radd__(self, other: Any) -> "Schema":
        if other == 0:
            return self.copy()
        other = self._coerce_other(other)
        return other.__add__(self)

    def __iadd__(self, other: Any) -> "Schema":
        other = self._coerce_other(other)

        for name, field in other.inner_fields.items():
            self.inner_fields[name] = field.copy()

        self.metadata = self._merge_metadata(self.metadata, other.metadata)
        return self

    def __sub__(self, other: Any) -> "Schema":
        other = self._coerce_other(other)
        remove_names = set(other.inner_fields)

        return Schema(
            inner_fields=OrderedDict(
                (name, field.copy())
                for name, field in self.inner_fields.items()
                if name not in remove_names
            ),
            metadata=dict(self.metadata) if self.metadata else None,
        )

    def __isub__(self, other: Any) -> "Schema":
        other = self._coerce_other(other)

        for name in tuple(other.inner_fields):
            self.inner_fields.pop(name, None)

        return self

    def __and__(self, other: Any) -> "Schema":
        other = self._coerce_other(other)
        keep_names = set(other.inner_fields)

        return Schema(
            inner_fields=OrderedDict(
                (name, field.copy())
                for name, field in self.inner_fields.items()
                if name in keep_names
            ),
            metadata=self._merge_metadata(self.metadata, other.metadata),
        )

    def __iand__(self, other: Any) -> "Schema":
        other = self._coerce_other(other)
        keep_names = set(other.inner_fields)

        self.inner_fields = OrderedDict(
            (name, field)
            for name, field in self.inner_fields.items()
            if name in keep_names
        )
        self.metadata = self._merge_metadata(self.metadata, other.metadata)
        return self

    def __or__(self, other: Any) -> "Schema":
        return self.__add__(other)

    def __ror__(self, other: Any) -> "Schema":
        other = self._coerce_other(other)
        return other.__add__(self)

    def __ior__(self, other: Any) -> "Schema":
        return self.__iadd__(other)

    def autotag(
        self,
        tags: dict[AnyStr, AnyStr] | None = None,
    ):
        f = OrderedDict()
        for name, field in self.inner_fields.items():
            f[name] = field.autotag()

        self.update_tags(tags)

        return Schema(
            inner_fields=f,
            metadata=self.metadata,
        )

    @classmethod
    def from_any(cls, obj: Any) -> "Schema":
        if isinstance(obj, cls):
            return obj

        from yggdrasil.arrow.cast import any_to_arrow_schema
        return cls.from_arrow(any_to_arrow_schema(obj))

    @classmethod
    def from_fields(
        cls,
        fields: Iterable[Field | pa.Field],
        *,
        metadata: dict[bytes | str, bytes | str | object] | None = None,
        tags: dict[bytes | str, bytes | str | object] | None = None,
    ) -> "Schema":
        inner_fields: OrderedDict[str, Field] = OrderedDict()

        for value in fields:
            field = Field.from_any(value)
            inner_fields[field.name] = field

        return cls(
            inner_fields=inner_fields,
            metadata=_normalize_metadata(metadata, tags=tags),
        )

    @classmethod
    def from_arrow(cls, value: pa.Schema) -> "Schema":
        if not isinstance(value, pa.Schema):
            from yggdrasil.arrow.cast import any_to_arrow_schema
            value = any_to_arrow_schema(value)

        inner_fields: OrderedDict[str, Field] = OrderedDict()
        for arrow_field in value:
            field = Field.from_arrow(arrow_field)
            inner_fields[field.name] = field

        return cls(
            inner_fields=inner_fields,
            metadata=dict(value.metadata) if value.metadata else None,
        )

    @classmethod
    def from_arrow_schema(cls, value: pa.Schema) -> "Schema":
        return cls.from_arrow(value)

    def to_arrow_schema(self) -> pa.Schema:
        return pa.schema(
            fields=[field.to_arrow_field() for field in self.inner_fields.values()],
            metadata=self.metadata,
        )

    @classmethod
    def from_polars(
        cls,
        obj: "pl.Schema | pl.DataFrame | pl.LazyFrame",
        *,
        metadata: dict[bytes | str, bytes | str | object] | None = None,
    ) -> "Schema":
        from yggdrasil.polars.lib import polars as pl

        if isinstance(obj, pl.Schema):
            fields = [
                Field.from_polars(name=name, dtype=dtype) for name, dtype in obj.items()
            ]
            return cls.from_fields(fields, metadata=metadata)

        if isinstance(obj, pl.DataFrame):
            fields = [
                Field.from_polars(name=name, dtype=dtype)
                for name, dtype in obj.schema.items()
            ]
            return cls.from_fields(fields, metadata=metadata)

        if isinstance(obj, pl.LazyFrame):
            fields = [
                Field.from_polars(name=name, dtype=dtype)
                for name, dtype in obj.collect_schema().items()
            ]
            return cls.from_fields(fields, metadata=metadata)

        raise TypeError(
            "Expected a polars.Schema, polars.DataFrame, or polars.LazyFrame; "
            f"got {type(obj).__name__}"
        )

    def to_polars_schema(self) -> "pl.Schema":
        from yggdrasil.polars.lib import polars as pl

        return pl.Schema(
            {
                field.name: field.to_polars_field().dtype
                for field in self.inner_fields.values()
            }
        )

    @classmethod
    def from_pyspark(
        cls,
        obj: "pst.StructType | Any",
        *,
        metadata: dict[bytes | str, bytes | str | object] | None = None,
        tags: dict[bytes | str, bytes | str | object] | None = None,
    ) -> "Schema":
        from yggdrasil.spark.cast import spark_schema_to_arrow_schema
        import pyspark.sql as pyspark_sql

        if isinstance(obj, pyspark_sql.types.StructType):
            arrow_schema = spark_schema_to_arrow_schema(obj)
            merged_metadata = (
                _normalize_metadata(metadata, tags=tags)
                if (metadata is not None or tags is not None)
                else (dict(arrow_schema.metadata) if arrow_schema.metadata else None)
            )
            return cls.from_fields(arrow_schema, metadata=merged_metadata)

        return cls.from_any(obj).copy(metadata=metadata, tags=tags)

    def to_pyspark_schema(self) -> "pst.StructType":
        from yggdrasil.spark.cast import arrow_schema_to_spark_schema

        return arrow_schema_to_spark_schema(self.to_arrow_schema())

    def cast_table(self, obj: Any, *, safe: bool = True) -> Any:
        from yggdrasil.data.cast import CastOptions
        return CastOptions(
            target_field=self.to_arrow_schema(),
            safe=safe,
        ).cast_table(obj)

    def cast_unstructured(
        self,
        obj: Any,
        *,
        as_type: type = pa.Table,
        safe: bool = True,
    ) -> Any:
        from yggdrasil.data.cast import CastOptions, convert
        options = CastOptions(target_field=self.to_arrow_schema(), safe=safe)
        return convert(obj, target_hint=as_type, options=options)



"""
AsyncJob — abstract awaitable handle for an async operation.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generic, Optional, TypeVar

from yggdrasil.dataclasses.waiting import WaitingConfigArg

from ..job_result import JobResult

__all__ = ["AsyncJob"]

T = TypeVar("T")


class AsyncJob(ABC, Generic[T]):
    """Abstract handle for an async operation that can be awaited.

    Implementors must provide :meth:`wait`, :attr:`is_done`, and :meth:`result`.
    """

    @abstractmethod
    def wait(
        self,
        wait: WaitingConfigArg = None,
        raise_error: bool = True,
    ) -> Optional[T]:
        """Block until the job completes and return its result.

        Args:
            wait:        ``None`` / ``True``  → block indefinitely.
                         ``False``            → non-blocking poll; returns ``None``
                         if not yet finished.
                         ``WaitingConfig`` or a numeric number of seconds → timed wait.
            raise_error: Re-raise the job's exception when ``True``; return
                         ``None`` when ``False``.

        Returns:
            The job's return value, or ``None`` on failure / timeout / non-blocking
            miss (when ``raise_error=False``).

        Raises:
            TimeoutError: If a timed wait elapses before the job finishes.
            BaseException: Any exception raised by the job (when ``raise_error=True``).
        """

    @property
    @abstractmethod
    def is_done(self) -> bool:
        """``True`` when the job has finished (success or failure)."""

    @abstractmethod
    def result(self) -> Optional[JobResult[T]]:
        """Return the :class:`~yggdrasil.concurrent.job_result.JobResult` if done, else ``None``."""




"""
Immutable result wrapper for an async job.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Generic, Optional, TypeVar

__all__ = ["JobResult"]

T = TypeVar("T")


@dataclass(frozen=True, slots=True)
class JobResult(Generic[T]):
    """Immutable outcome of a :class:`~yggdrasil.concurrent.job.Job` execution.

    Either holds a *result* (success) or an *exception* (failure).
    Both fields are always present; exactly one is ``None`` in the normal case.
    """

    result: Optional[T]
    exception: Optional[BaseException]

    # ------------------------------------------------------------------
    # Convenience helpers
    # ------------------------------------------------------------------

    @property
    def ok(self) -> bool:
        """``True`` when the job succeeded (no exception captured)."""
        return self.exception is None

    def get(self) -> T:
        """Return the result or re-raise the captured exception.

        Raises:
            BaseException: The exception that was captured during job execution.
        """
        if self.exception is not None:
            raise self.exception
        return self.result  # type: ignore[return-value]

    def __bool__(self) -> bool:
        return self.ok




"""Browser-like HTTP session that simulates a real browser client.

``BrowserHTTPSession`` extends :class:`~yggdrasil.io.http_.session.HTTPSession`
with everything needed to impersonate a browser:

* Realistic ``User-Agent`` strings via :class:`~yggdrasil.io.user_agents.UserAgentGenerator`
* Full browser header suite per request (``Accept``, ``Accept-Language``,
  ``Accept-Encoding``, ``Cache-Control``, ``Upgrade-Insecure-Requests``,
  ``Sec-Fetch-*``, ``sec-ch-ua``, ``sec-ch-ua-mobile``, ``sec-ch-ua-platform``)
* Persistent cookie jar with automatic ``Set-Cookie`` parsing
* Referrer tracking and ``Sec-Fetch-Site`` computation
* Navigation helpers: :meth:`navigate`, :meth:`follow_link`, :meth:`submit_form`
* Convenience wrappers: :meth:`get`, :meth:`post`, :meth:`head`
* User-agent presets via :meth:`set_browser_preset` and rotation via
  :meth:`rotate_user_agent`
* Browser / platform introspection: :attr:`browser_name`, :attr:`platform`,
  :attr:`is_mobile`, :attr:`sec_ch_ua`, :attr:`sec_ch_ua_platform`
"""
from __future__ import annotations

import re
from dataclasses import dataclass, field
from typing import Any, Mapping, Optional
from urllib.parse import urlencode, urlsplit

from yggdrasil.io.request import PreparedRequest
from yggdrasil.io.send_config import SendConfig
from yggdrasil.io.url import URL
from yggdrasil.io.user_agents import UserAgentGenerator

from .response import HTTPResponse
from .session import HTTPSession

__all__ = ["BrowserHTTPSession"]

# ---------------------------------------------------------------------------
# Constants
# ---------------------------------------------------------------------------

_BROWSER_ACCEPT = (
    "text/html,application/xhtml+xml,application/xml;q=0.9,"
    "image/avif,image/webp,image/apng,*/*;q=0.8,"
    "application/signed-exchange;v=b3;q=0.7"
)

_BROWSER_ACCEPT_ENCODING = "gzip, deflate, br, zstd"

# User-agent pattern regexes
_CHROME_RE = re.compile(r"Chrome/(\d+)\.")
_EDGE_RE = re.compile(r"Edg(?:e)?/(\d+)\.")
_FIREFOX_RE = re.compile(r"Firefox/(\d+)\.")
_SAFARI_RE = re.compile(r"\bSafari/")
_WINDOWS_RE = re.compile(r"Windows NT")
_MAC_RE = re.compile(r"Macintosh|Mac OS X")
_LINUX_RE = re.compile(r"Linux")
_ANDROID_RE = re.compile(r"Android")
_IOS_RE = re.compile(r"iPhone|iPad|iPod")
_MOBILE_RE = re.compile(r"Mobile|Android")


# ---------------------------------------------------------------------------
# BrowserHTTPSession
# ---------------------------------------------------------------------------


@dataclass
class BrowserHTTPSession(HTTPSession):
    """An :class:`~yggdrasil.io.http_.session.HTTPSession` that impersonates
    a real browser client.

    Features
    --------
    * Realistic ``User-Agent`` strings generated by
      :class:`~yggdrasil.io.user_agents.UserAgentGenerator`.
    * Full browser header suite sent with every request: ``Accept``,
      ``Accept-Language``, ``Accept-Encoding``, ``Cache-Control``,
      ``Upgrade-Insecure-Requests``, ``Sec-Fetch-*``, and Chromium client-hint
      headers (``sec-ch-ua``, ``sec-ch-ua-mobile``, ``sec-ch-ua-platform``).
    * Persistent cookie jar with automatic ``Set-Cookie`` parsing.
    * Referrer tracking and ``Sec-Fetch-Site`` computation.
    * Navigation helpers: :meth:`navigate`, :meth:`follow_link`,
      :meth:`submit_form`, plus :meth:`get`, :meth:`post`, :meth:`head`
      convenience wrappers.
    * User-agent management: :meth:`set_user_agent`, :meth:`rotate_user_agent`,
      :meth:`set_browser_preset`.
    * Browser / platform introspection: :attr:`browser_name`, :attr:`platform`,
      :attr:`is_mobile`, :attr:`sec_ch_ua`, :attr:`sec_ch_ua_platform`.

    Quick start::

        session = HTTPSession(base_url="https://example.com")
        browser = session.to_browser()
        resp = browser.navigate("https://example.com/login")
        resp = browser.submit_form(
            "https://example.com/login",
            {"username": "alice", "password": "secret"},
        )

    Or construct directly::

        browser = BrowserHTTPSession(base_url="https://example.com")
        browser.set_browser_preset("firefox", platform="linux")
    """

    # ── Configuration fields ──────────────────────────────────────────────────

    user_agent: Optional[str] = field(default=None)
    """Explicit ``User-Agent`` string.  ``None`` (default) = auto-generate."""

    accept_language: str = field(default="en-US,en;q=0.9")
    """Value for the ``Accept-Language`` request header."""

    accept: str = field(default=_BROWSER_ACCEPT)
    """Value for the ``Accept`` request header."""

    ua_seed: Optional[int] = field(default=None)
    """Optional RNG seed passed to :class:`UserAgentGenerator` for
    deterministic user-agent generation."""

    # ── Runtime state ─────────────────────────────────────────────────────────

    _cookies: dict[str, str] = field(
        default_factory=dict, init=False, repr=False, compare=False
    )
    _referrer: Optional[str] = field(
        default=None, init=False, repr=False, compare=False
    )
    _ua_generator: Optional[UserAgentGenerator] = field(
        default=None, init=False, repr=False, compare=False
    )

    # ── Lifecycle ─────────────────────────────────────────────────────────────

    def __post_init__(self) -> None:
        self._ua_generator = UserAgentGenerator(seed=self.ua_seed)
        if self.user_agent is None:
            self.user_agent = self._ua_generator.random()
        super().__post_init__()

    # ── UA introspection ──────────────────────────────────────────────────────

    @property
    def is_mobile(self) -> bool:
        """``True`` when the active user-agent string looks like a mobile
        browser (Android / iPhone / iPad / Mobile token)."""
        return bool(_MOBILE_RE.search(self.user_agent or ""))

    @property
    def platform(self) -> str:
        """Operating-system name detected from the active user-agent.

        Returns one of ``"Windows"``, ``"macOS"``, ``"Linux"``,
        ``"Android"``, ``"iOS"``, or ``"Unknown"``.
        """
        ua = self.user_agent or ""
        if _ANDROID_RE.search(ua):
            return "Android"
        if _IOS_RE.search(ua):
            return "iOS"
        if _WINDOWS_RE.search(ua):
            return "Windows"
        if _MAC_RE.search(ua):
            return "macOS"
        if _LINUX_RE.search(ua):
            return "Linux"
        return "Unknown"

    @property
    def browser_name(self) -> str:
        """Best-effort browser name detected from the active user-agent string.

        Returns one of ``"Edge"``, ``"Chrome"``, ``"Firefox"``, ``"Safari"``,
        or ``"Unknown"``.
        """
        ua = self.user_agent or ""
        if _EDGE_RE.search(ua):
            return "Edge"
        if _CHROME_RE.search(ua):
            return "Chrome"
        if _FIREFOX_RE.search(ua):
            return "Firefox"
        if _SAFARI_RE.search(ua):
            return "Safari"
        return "Unknown"

    @property
    def sec_ch_ua(self) -> Optional[str]:
        """Chromium-style ``sec-ch-ua`` header derived from the active UA.

        Returns ``None`` for non-Chromium browsers (Firefox, Safari) because
        they do not send this header.

        Examples
        --------
        Chrome 124::

            '"Google Chrome";v="124", "Chromium";v="124", "Not-A.Brand";v="99"'

        Edge 124::

            '"Microsoft Edge";v="124", "Chromium";v="124", "Not-A.Brand";v="99"'
        """
        ua = self.user_agent or ""
        edge_m = _EDGE_RE.search(ua)
        chrome_m = _CHROME_RE.search(ua)
        if edge_m:
            v = edge_m.group(1)
            return (
                f'"Microsoft Edge";v="{v}", '
                f'"Chromium";v="{v}", '
                '"Not-A.Brand";v="99"'
            )
        if chrome_m:
            v = chrome_m.group(1)
            return (
                f'"Google Chrome";v="{v}", '
                f'"Chromium";v="{v}", '
                '"Not-A.Brand";v="99"'
            )
        return None

    @property
    def sec_ch_ua_platform(self) -> str:
        """Platform string used in the ``sec-ch-ua-platform`` client-hint
        header (e.g. ``"Windows"``, ``"macOS"``, ``"Linux"``)."""
        _map = {
            "Windows": "Windows",
            "macOS": "macOS",
            "Linux": "Linux",
            "Android": "Android",
            "iOS": "iOS",
        }
        return _map.get(self.platform, "Unknown")

    # ── Cookie jar ────────────────────────────────────────────────────────────

    @property
    def cookies(self) -> dict[str, str]:
        """Snapshot of the current session cookie jar (name → value)."""
        return dict(self._cookies)

    def set_cookie(self, name: str, value: str) -> None:
        """Add or update *name* in the session cookie jar."""
        self._cookies[name] = value

    def delete_cookie(self, name: str) -> None:
        """Remove *name* from the session cookie jar (no-op if absent)."""
        self._cookies.pop(name, None)

    def clear_cookies(self) -> None:
        """Empty the session cookie jar."""
        self._cookies.clear()

    def _serialize_cookies(self) -> str:
        """Serialize the cookie jar into a ``Cookie`` header value."""
        return "; ".join(f"{k}={v}" for k, v in self._cookies.items())

    @staticmethod
    def _parse_set_cookie(header: str) -> tuple[str, str]:
        """Parse a single ``Set-Cookie`` header value into ``(name, value)``.

        Attributes (``Path``, ``Expires``, ``HttpOnly``, ``Secure``, …) are
        intentionally ignored for simplicity.
        """
        name_value, _, _ = header.partition(";")
        name_value = name_value.strip()
        if "=" in name_value:
            name, _, value = name_value.partition("=")
            return name.strip(), value.strip()
        return name_value, ""

    def _update_cookies_from_response(self, response: HTTPResponse) -> None:
        """Extract ``Set-Cookie`` headers from *response* and merge them into
        the cookie jar.

        urllib3 may collapse multiple ``Set-Cookie`` values with a comma, so
        this method splits on ``","`` as a best-effort fallback.
        """
        raw = (
            response.headers.get("Set-Cookie")
            or response.headers.get("set-cookie")
        )
        if not raw:
            return
        for cookie_header in raw.split(","):
            try:
                name, value = self._parse_set_cookie(cookie_header.strip())
                if name:
                    self._cookies[name] = value
            except Exception:
                continue

    # ── Referrer ──────────────────────────────────────────────────────────────

    @property
    def referrer(self) -> Optional[str]:
        """URL of the last successfully navigated page, sent as the
        ``Referer`` header on subsequent requests."""
        return self._referrer

    def set_referrer(self, url: str) -> None:
        """Manually set the ``Referer`` header for subsequent requests."""
        self._referrer = str(url) if url else None

    def clear_referrer(self) -> None:
        """Reset the referrer so the next request carries no ``Referer``
        header."""
        self._referrer = None

    # ── User-agent management ─────────────────────────────────────────────────

    def set_user_agent(self, ua: str) -> None:
        """Explicitly set the ``User-Agent`` string."""
        self.user_agent = str(ua)

    def rotate_user_agent(self, seed: Optional[int] = None) -> str:
        """Generate and apply a new random ``User-Agent`` string.

        Parameters
        ----------
        seed:
            Optional integer seed for deterministic generation.

        Returns
        -------
        str
            The newly applied user-agent string.
        """
        self._ua_generator = UserAgentGenerator(seed=seed)
        self.user_agent = self._ua_generator.random()
        return self.user_agent

    def set_browser_preset(
        self,
        browser: str = "chrome",
        *,
        platform: str = "windows",
        seed: Optional[int] = None,
    ) -> str:
        """Set ``User-Agent`` to a plausible UA for a specific
        *browser* / *platform* combination.

        The method samples :class:`~yggdrasil.io.user_agents.UserAgentGenerator`
        up to 20 times until a UA matching the requested browser is found.

        Parameters
        ----------
        browser:
            ``"chrome"``, ``"edge"``, ``"firefox"``, or ``"safari"``.
        platform:
            ``"windows"``, ``"mac"``, ``"linux"``, ``"android"``, or
            ``"ios"``.
        seed:
            Optional base integer seed for reproducibility.

        Returns
        -------
        str
            The applied user-agent string.

        Examples
        --------
        ::

            browser.set_browser_preset("firefox", platform="linux")
            browser.set_browser_preset("edge", platform="windows", seed=42)
        """
        from random import Random

        browser_l = browser.lower()
        platform_l = platform.lower()

        _platform_method_map: dict[str, str] = {
            "windows": "_desktop_windows",
            "mac": "_desktop_mac",
            "linux": "_desktop_linux",
            "android": "_android",
            "ios": "_ios",
        }
        method_name = _platform_method_map.get(platform_l, "_desktop_windows")

        ua: str = ""
        for attempt in range(20):
            current_seed = (seed + attempt) if seed is not None else attempt
            gen = UserAgentGenerator(seed=current_seed)
            r = Random(current_seed)
            ua = getattr(gen, method_name)(r)
            ua_lower = ua.lower()

            if browser_l == "edge" and "edg/" in ua_lower:
                self.user_agent = ua
                return ua
            if (
                browser_l in ("chrome", "chromium")
                and "chrome/" in ua_lower
                and "edg/" not in ua_lower
            ):
                self.user_agent = ua
                return ua
            if browser_l == "firefox" and "firefox/" in ua_lower:
                self.user_agent = ua
                return ua
            if (
                browser_l == "safari"
                and "safari/" in ua_lower
                and "chrome/" not in ua_lower
                and "firefox/" not in ua_lower
            ):
                self.user_agent = ua
                return ua

        # Fallback: apply the last generated UA regardless of browser match
        self.user_agent = ua
        return ua

    # ── Header construction ───────────────────────────────────────────────────

    def _compute_sec_fetch_site(self, request: PreparedRequest) -> str:
        """Compute the ``Sec-Fetch-Site`` value by comparing the active
        referrer against the target URL.

        Returns one of ``"none"`` (no referrer), ``"same-origin"``,
        ``"same-site"``, or ``"cross-site"``.

        The same-site check uses a two-label registrable-domain heuristic
        (e.g. ``www.example.com`` and ``api.example.com`` are same-site
        because both share the ``example.com`` registrable domain).
        """
        if not self._referrer:
            return "none"
        try:
            ref = urlsplit(self._referrer)
            req = urlsplit(request.url.to_string())
            if ref.netloc == req.netloc:
                return "same-origin"
            ref_host = ref.hostname or ""
            req_host = req.hostname or ""
            if ref_host and req_host:
                if self._registrable_domain(ref_host) == self._registrable_domain(req_host):
                    return "same-site"
            return "cross-site"
        except Exception:
            return "none"

    @staticmethod
    def _registrable_domain(host: str) -> str:
        """Return a best-effort registrable domain (last two DNS labels).

        This is a lightweight heuristic — it does not consult the public
        suffix list.  Sufficient for ``Sec-Fetch-Site`` simulation in the
        common single-TLD case (``example.com``, ``example.co``, …).
        """
        parts = host.rsplit(".", 2)
        return ".".join(parts[-2:]) if len(parts) >= 2 else host


    def _browser_headers_for_request(
        self, request: PreparedRequest
    ) -> dict[str, str]:
        """Build the complete browser header dictionary for *request*.

        This includes ``User-Agent``, ``Accept``, ``Accept-Language``,
        ``Accept-Encoding``, ``Cache-Control``, ``Connection``,
        ``Upgrade-Insecure-Requests``, all ``Sec-Fetch-*`` headers,
        Chromium client-hint headers (when applicable), ``Cookie``, and
        ``Referer``.
        """
        headers: dict[str, str] = {
            "User-Agent": self.user_agent or "",
            "Accept": self.accept,
            "Accept-Language": self.accept_language,
            "Accept-Encoding": _BROWSER_ACCEPT_ENCODING,
            "Connection": "keep-alive",
            "Upgrade-Insecure-Requests": "1",
            "Cache-Control": "max-age=0",
            "Sec-Fetch-Dest": "document",
            "Sec-Fetch-Mode": "navigate",
            "Sec-Fetch-Site": self._compute_sec_fetch_site(request),
            "Sec-Fetch-User": "?1",
        }

        # Chromium UA client-hint headers (Chrome / Edge only)
        ch_ua = self.sec_ch_ua
        if ch_ua:
            headers["sec-ch-ua"] = ch_ua
            headers["sec-ch-ua-mobile"] = "?1" if self.is_mobile else "?0"
            headers["sec-ch-ua-platform"] = f'"{self.sec_ch_ua_platform}"'

        # Cookies
        if self._cookies:
            headers["Cookie"] = self._serialize_cookies()

        # Referrer
        if self._referrer:
            headers["Referer"] = self._referrer

        return headers

    def _build_request_headers(
        self,
        request: PreparedRequest,
    ) -> dict[str, str]:
        """Return the merged header dict for *request*.

        Browser defaults are set first; user-supplied :attr:`send_headers`
        override them so that explicit customisation always takes precedence.
        """
        headers = self._browser_headers_for_request(request)
        if self.send_headers:
            # User-set headers win over browser defaults
            headers.update(self.send_headers)
        return headers

    # ── URL resolution ────────────────────────────────────────────────────────

    def _resolve_url(self, url: URL | str) -> URL:
        """Resolve *url* to an absolute :class:`~yggdrasil.io.url.URL`.

        Resolution rules (evaluated in order):

        1. **Absolute** ``URL`` objects (has scheme + host) → returned unchanged.
        2. **Relative** ``URL`` objects → joined against :attr:`base_url`.
        3. ``https://…`` / ``http://…`` strings → parsed and returned.
        4. ``//…`` (protocol-relative) → ``https:`` prepended, then parsed.
        5. Starts with ``/``, ``./``, or ``../`` → absolute/relative path,
           joined against :attr:`base_url`.
        6. Windows drive-letter pattern (e.g. ``C:/path``) → :exc:`ValueError`.
        7. First segment before ``/`` contains ``"."`` → schemaless hostname
           (e.g. ``"example.com/path"``), ``https://`` prepended.
        8. Otherwise → relative path segment (e.g. ``"api/resource"``), joined
           against :attr:`base_url`.  :exc:`ValueError` if no :attr:`base_url`.

        Parameters
        ----------
        url:
            Raw URL string or :class:`~yggdrasil.io.url.URL` object.

        Returns
        -------
        URL
            Fully resolved, absolute URL object.

        Raises
        ------
        ValueError
            For Windows local paths or relative paths when no :attr:`base_url`
            is configured.
        """
        # ── URL objects ────────────────────────────────────────────────────
        if isinstance(url, URL):
            if url.is_absolute:
                return url
            # Relative URL object — join with base_url
            if self.base_url:
                return self.base_url.join(url.to_string())
            return url

        s = str(url).strip()
        if not s:
            raise ValueError("URL must not be empty.")

        # ── Already has explicit http/https ────────────────────────────────
        if s.startswith(("https://", "http://")):
            return URL.parse(s)

        # ── Protocol-relative ──────────────────────────────────────────────
        if s.startswith("//"):
            return URL.parse("https:" + s)

        # ── Explicit relative paths (/path, ./path, ../path) ──────────────
        if s.startswith(("/", "./", "../")):
            if self.base_url:
                return self.base_url.join(s)
            raise ValueError(
                f"Cannot resolve path-relative URL {s!r}: no base_url is set on "
                "this session. Set base_url or pass a full URL."
            )

        # ── Windows drive-letter detection (e.g. "C:/…", "D:\\…") ─────────
        if len(s) >= 3 and s[0].isalpha() and s[1] == ":" and s[2] in ("/", "\\"):
            raise ValueError(
                f"URL looks like a local Windows path, not an HTTP URL: {s!r}. "
                "Pass a full URL with scheme (e.g. 'https://host/path')."
            )

        # ── No explicit scheme — hostname vs relative path heuristic ───────
        # If the first segment (before the first "/") contains a "." that is
        # not a leading dot, treat it as a schemaless hostname.
        # Examples:
        #   "example.com/path"      → first_seg="example.com"  → has dot → hostname
        #   "api.example.com/v1"    → first_seg="api.example.com" → has dot → hostname
        #   "api/controlador.cgi"   → first_seg="api" → no dot → relative path
        #   "v1/resource"           → first_seg="v1"  → no dot → relative path
        first_seg = s.split("/", 1)[0] if "/" in s else s
        if "." in first_seg and not first_seg.startswith("."):
            return URL.parse("https://" + s)

        # ── Relative path segment (e.g. "api/resource", "page") ───────────
        if self.base_url:
            return self.base_url.join(s)

        raise ValueError(
            f"Cannot resolve relative URL {s!r}: no base_url is set on this "
            "session. Set base_url (e.g. HTTPSession(base_url='https://api.example.com')) "
            "or pass a full URL."
        )

    @staticmethod
    def _apply_params(url: URL, params: Mapping[str, Any]) -> URL:
        """Append *params* as query-string key/value pairs to *url*.

        Existing query parameters on *url* are preserved.  Multi-valued
        parameters can be expressed as a list or tuple.

        Parameters
        ----------
        url:
            Target URL.
        params:
            Mapping of query parameter names to values (scalar or sequence).

        Returns
        -------
        URL
            New URL with the extra query parameters appended.
        """
        for key, value in params.items():
            key_s = str(key)
            if isinstance(value, (list, tuple)):
                for v in value:
                    url = url.add_query_item(key_s, str(v), replace=False)
            else:
                url = url.add_query_item(key_s, str(value), replace=False)
        return url

    def get(
        self,
        url: URL | str,
        *,
        params: Optional[Mapping[str, Any]] = None,
        headers: Optional[Mapping[str, str]] = None,
        config: SendConfig | Mapping[str, Any] | None = None,
        **send_kwargs: Any,
    ) -> HTTPResponse:
        """Issue a browser-like ``GET`` request to *url*.

        Parameters
        ----------
        url:
            Destination URL.  Relative paths (``"api/resource"``) are joined
            against :attr:`base_url`.  Schemaless host strings
            (``"example.com/path"``) get ``https://`` prepended.
        params:
            Optional query-string parameters appended to the URL.
        headers:
            Optional per-request headers merged on top of browser defaults.
        config:
            :class:`~yggdrasil.io.send_config.SendConfig` or compatible dict.

        Returns
        -------
        HTTPResponse
        """
        resolved = self._resolve_url(url)
        if params:
            resolved = self._apply_params(resolved, params)
        request = PreparedRequest.prepare("GET", resolved, headers=headers)
        return self.send(request, config=config, **send_kwargs)  # type: ignore[return-value]

    def head(
        self,
        url: URL | str,
        *,
        params: Optional[Mapping[str, Any]] = None,
        headers: Optional[Mapping[str, str]] = None,
        config: SendConfig | Mapping[str, Any] | None = None,
        **send_kwargs: Any,
    ) -> HTTPResponse:
        """Issue a browser-like ``HEAD`` request to *url*.

        Parameters
        ----------
        url:
            Destination URL.  Relative paths are joined against :attr:`base_url`.
        params:
            Optional query-string parameters appended to the URL.
        headers:
            Optional per-request headers merged on top of browser defaults.
        config:
            :class:`~yggdrasil.io.send_config.SendConfig` or compatible dict.

        Returns
        -------
        HTTPResponse
        """
        resolved = self._resolve_url(url)
        if params:
            resolved = self._apply_params(resolved, params)
        request = PreparedRequest.prepare("HEAD", resolved, headers=headers)
        return self.send(request, config=config, **send_kwargs)  # type: ignore[return-value]

    def post(
        self,
        url: URL | str,
        *,
        params: Optional[Mapping[str, Any]] = None,
        body: Optional[Any] = None,
        json: Optional[Any] = None,
        headers: Optional[Mapping[str, str]] = None,
        config: SendConfig | Mapping[str, Any] | None = None,
        **send_kwargs: Any,
    ) -> HTTPResponse:
        """Issue a browser-like ``POST`` request to *url*.

        Parameters
        ----------
        url:
            Destination URL.  Relative paths are joined against :attr:`base_url`.
        params:
            Optional query-string parameters appended to the URL.
        body:
            Raw request body.  Mutually exclusive with *json*.
        json:
            JSON-serialisable body.  Sets ``Content-Type: application/json``.
        headers:
            Optional per-request headers merged on top of browser defaults.
        config:
            :class:`~yggdrasil.io.send_config.SendConfig` or compatible dict.

        Returns
        -------
        HTTPResponse
        """
        resolved = self._resolve_url(url)
        if params:
            resolved = self._apply_params(resolved, params)
        request = PreparedRequest.prepare(
            "POST", resolved, headers=headers, body=body, json=json
        )
        return self.send(request, config=config, **send_kwargs)  # type: ignore[return-value]

    # ── Navigation helpers ────────────────────────────────────────────────────

    def navigate(
        self,
        url: URL | str,
        *,
        method: str = "GET",
        params: Optional[Mapping[str, Any]] = None,
        body: Optional[Any] = None,
        json: Optional[Any] = None,
        headers: Optional[Mapping[str, str]] = None,
        update_referrer: bool = True,
        update_cookies: bool = True,
        raise_error: bool = True,
        config: SendConfig | Mapping[str, Any] | None = None,
        **send_kwargs: Any,
    ) -> HTTPResponse:
        """Navigate to *url*, optionally updating the referrer and cookie jar.

        The response is returned *after* updating session state so that
        subsequent requests benefit from the new referrer / cookies.

        Parameters
        ----------
        url:
            Destination URL (string or :class:`~yggdrasil.io.url.URL`).
            Relative paths are joined against :attr:`base_url`.
        method:
            HTTP method (default ``"GET"``).
        params:
            Optional query-string parameters appended to the URL.
        body:
            Raw request body.  Mutually exclusive with *json*.
        json:
            JSON-serialisable body.  Sets ``Content-Type: application/json``.
        headers:
            Optional per-request headers merged on top of browser defaults.
        update_referrer:
            When ``True`` (default) and the response is ``< 400``, update
            :attr:`referrer` to the request URL so the next request carries
            the correct ``Referer`` / ``Sec-Fetch-Site``.
        update_cookies:
            When ``True`` (default), extract and store ``Set-Cookie`` headers
            from the response automatically.
        raise_error:
            When ``True`` (default), call :meth:`~HTTPResponse.raise_for_status`
            after updating session state.
        config:
            :class:`~yggdrasil.io.send_config.SendConfig` or compatible dict.

        Returns
        -------
        HTTPResponse
        """
        resolved = self._resolve_url(url)
        if params:
            resolved = self._apply_params(resolved, params)
        request = PreparedRequest.prepare(
            method, resolved, headers=headers, body=body, json=json
        )
        # Send with raise_error=False so we can update state before raising
        result: HTTPResponse = self.send(  # type: ignore[assignment]
            request, config=config, raise_error=False, **send_kwargs
        )

        if update_cookies:
            self._update_cookies_from_response(result)
        if update_referrer and result.status_code < 400:
            self._referrer = request.url.to_string()

        if raise_error:
            result.raise_for_status()

        return result

    def follow_link(
        self,
        url: URL | str,
        *,
        from_url: Optional[str] = None,
        headers: Optional[Mapping[str, str]] = None,
        raise_error: bool = True,
        config: SendConfig | Mapping[str, Any] | None = None,
        **send_kwargs: Any,
    ) -> HTTPResponse:
        """Simulate clicking a hyperlink.

        Sets *from_url* (or the current :attr:`referrer`) as the ``Referer``
        before navigating, then updates :attr:`referrer` to *url* on a
        successful response.

        Parameters
        ----------
        url:
            The link destination.  Relative paths are joined against
            :attr:`base_url`.
        from_url:
            Explicit referrer to use before this navigation; when ``None``
            (default) the current :attr:`referrer` is kept.
        headers:
            Optional per-request headers merged on top of browser defaults.
        raise_error:
            Raise on HTTP error responses (default ``True``).
        config:
            :class:`~yggdrasil.io.send_config.SendConfig` or compatible dict.

        Returns
        -------
        HTTPResponse
        """
        if from_url is not None:
            self._referrer = from_url
        return self.navigate(
            url,
            method="GET",
            headers=headers,
            raise_error=raise_error,
            config=config,
            **send_kwargs,
        )

    def submit_form(
        self,
        url: URL | str,
        form_data: Mapping[str, Any],
        *,
        method: str = "POST",
        extra_headers: Optional[Mapping[str, str]] = None,
        raise_error: bool = True,
        config: SendConfig | Mapping[str, Any] | None = None,
        **send_kwargs: Any,
    ) -> HTTPResponse:
        """Simulate an HTML form submission.

        *form_data* is URL-encoded (``application/x-www-form-urlencoded``)
        and sent as the request body.  ``Set-Cookie`` headers from the
        response are automatically merged into the cookie jar.

        Parameters
        ----------
        url:
            Form action URL.  Relative paths are joined against :attr:`base_url`.
        form_data:
            Mapping of field names to values.
        method:
            HTTP method (default ``"POST"``).
        extra_headers:
            Additional headers merged on top of browser defaults.
        raise_error:
            Raise on HTTP error responses (default ``True``).
        config:
            :class:`~yggdrasil.io.send_config.SendConfig` or compatible dict.

        Returns
        -------
        HTTPResponse
        """
        encoded = urlencode(
            {str(k): str(v) for k, v in form_data.items()}
        ).encode("utf-8")

        merged_headers: dict[str, str] = {
            "Content-Type": "application/x-www-form-urlencoded",
        }
        if extra_headers:
            merged_headers.update(extra_headers)

        request = PreparedRequest.prepare(
            method, self._resolve_url(url), headers=merged_headers, body=encoded
        )
        result: HTTPResponse = self.send(  # type: ignore[assignment]
            request, config=config, raise_error=False, **send_kwargs
        )

        self._update_cookies_from_response(result)

        if raise_error:
            result.raise_for_status()

        return result






"""
ThreadJob — daemon-thread-backed implementation of AsyncJob.
"""

from __future__ import annotations

import logging
import threading as _threading
from dataclasses import dataclass, field
from threading import Thread
from typing import Optional, TypeVar

from yggdrasil.dataclasses.waiting import WaitingConfig, WaitingConfigArg

from ..job import Job
from ..job_result import JobResult
from .base import AsyncJob

__all__ = ["ThreadJob"]

LOGGER = logging.getLogger(__name__)

T = TypeVar("T")


@dataclass(slots=True)
class ThreadJob(AsyncJob[T]):
    """A :class:`~yggdrasil.concurrent.job.Job` running in a daemon
    :class:`~threading.Thread`, started immediately on construction.

    Use :meth:`wait` to block and retrieve the result::

        handle = Job.make(fetch_data, url).fire_and_forget()

        # … do other work …

        data = handle.wait()           # block until thread finishes
        data = handle.wait(wait=5.0)   # block at most 5 seconds
        data = handle.wait(wait=False) # non-blocking poll
    """

    job: Job[T]

    # Internal mutable state — excluded from the public constructor
    _result: Optional[T] = field(default=None, init=False, repr=False)
    _exception: Optional[BaseException] = field(default=None, init=False, repr=False)
    _done: _threading.Event = field(
        default_factory=_threading.Event, init=False, repr=False,
    )
    _thread: Optional[Thread] = field(default=None, init=False, repr=False)

    def __post_init__(self) -> None:
        func_name = getattr(self.job.func, "__qualname__", repr(self.job.func))
        t = Thread(
            target=self._run,
            name=f"ThreadJob-{func_name}",
            daemon=True,
        )
        self._thread = t
        t.start()
        LOGGER.debug("ThreadJob started: %s", func_name)

    # ------------------------------------------------------------------
    # Thread body
    # ------------------------------------------------------------------

    def _run(self) -> None:
        func_name = getattr(self.job.func, "__qualname__", repr(self.job.func))
        try:
            self._result = self.job.run()
            LOGGER.debug("ThreadJob finished ok: %s", func_name)
        except BaseException as exc:  # noqa: BLE001
            self._exception = exc
            LOGGER.debug("ThreadJob raised %r: %s", exc, func_name)
        finally:
            self._done.set()

    # ------------------------------------------------------------------
    # AsyncJob implementation
    # ------------------------------------------------------------------

    @property
    def is_done(self) -> bool:
        """``True`` when the thread has finished (success or failure)."""
        return self._done.is_set()

    def result(self) -> Optional[JobResult[T]]:
        """Return the :class:`~yggdrasil.concurrent.job_result.JobResult` if done, else ``None``."""
        if not self.is_done:
            return None
        return JobResult(result=self._result, exception=self._exception)

    def wait(
        self,
        wait: WaitingConfigArg = None,
        raise_error: bool = True,
    ) -> Optional[T]:
        """Block until the thread finishes.

        Args:
            wait:        ``None`` / ``True``  → block indefinitely.
                         ``False``            → non-blocking poll.
                         ``WaitingConfig`` or numeric seconds → timed wait.
            raise_error: Re-raise the thread's exception when ``True``.

        Returns:
            The job return value, or ``None`` on failure / timeout / non-blocking
            miss (when ``raise_error=False``).

        Raises:
            TimeoutError: When a timed wait elapses before the thread finishes.
            BaseException: The exception from the job (when ``raise_error=True``).
        """
        # --- Non-blocking poll ---
        if wait is False:
            if not self._done.is_set():
                return None

        # --- Block indefinitely ---
        elif wait is None or wait is True:
            self._done.wait(timeout=None)

        # --- Timed wait via WaitingConfig ---
        else:
            config = WaitingConfig.check_arg(wait)
            timeout: Optional[float] = config.timeout if config.timeout else None
            completed = self._done.wait(timeout=timeout)
            if not completed:
                func_name = getattr(self.job.func, "__qualname__", repr(self.job.func))
                LOGGER.warning(
                    "ThreadJob timed out after %.1fs: %s", timeout, func_name,
                )
                if raise_error:
                    raise TimeoutError(
                        f"ThreadJob({func_name!r}) did not complete within {timeout}s"
                    )
                return None

        # --- Propagate exception or return result ---
        if self._exception is not None:
            if raise_error:
                raise self._exception
            return None

        return self._result




"""
yggdrasil.concurrent.threading
==============================
Thread-specific async job primitives.

Canonical exports
-----------------
- :class:`AsyncJob`  — abstract awaitable handle  (``base.py``)
- :class:`ThreadJob` — daemon-thread implementation (``thread_job.py``)

Backward-compatible re-exports
-------------------------------
``Job``, ``JobResult``, and ``JobPoolExecutor`` are also exported here so that
existing ``from yggdrasil.concurrent.threading import …`` statements continue
to work without modification.
"""

from __future__ import annotations

from .base import AsyncJob
from .thread_job import ThreadJob

# Backward-compat re-exports
from ..job import Job
from ..job_result import JobResult
from ..pool import JobPoolExecutor

__all__ = [
    # Thread-specific (canonical)
    "AsyncJob",
    "ThreadJob",
    # Backward-compat re-exports
    "Job",
    "JobResult",
    "JobPoolExecutor",
]




from __future__ import annotations

from typing import Any

from bson import json_util


class BsonJsonSerde:
    """Serialize BSON-capable Python objects through MongoDB Extended JSON."""

    @staticmethod
    def dumps(value: Any) -> str:
        return json_util.dumps(value)

    @staticmethod
    def loads(payload: str) -> Any:
        return json_util.loads(payload)

    @staticmethod
    def normalize(value: Any) -> Any:
        return json_util.loads(json_util.dumps(value))



from __future__ import annotations

from typing import Any, Dict, Iterator, List, Optional, Sequence, Tuple

from .cursor import HttpMongoCursor
from .results import DeleteResult, InsertManyResult, InsertOneResult, UpdateResult
from .transport import HttpTransport


class HttpCollection:
    def __init__(self, transport: HttpTransport, database: str, name: str) -> None:
        self._transport = transport
        self.database = database
        self.name = name
        self.full_name = f"{database}.{name}"

    def _rpc(self, operation: str, arguments: Dict[str, Any]) -> Any:
        return self._transport.rpc(
            database=self.database,
            collection=self.name,
            operation=operation,
            arguments=arguments,
        )

    def find(
        self,
        filter: Optional[Dict[str, Any]] = None,
        projection: Optional[Dict[str, Any]] = None,
        skip: int = 0,
        limit: int = 0,
        sort: Optional[List[Tuple[str, int]]] = None,
        **kwargs: Any,
    ) -> HttpMongoCursor:
        documents = self._rpc(
            "find",
            {
                "filter": filter or {},
                "projection": projection,
                "skip": skip,
                "limit": limit,
                "sort": sort,
                "kwargs": kwargs,
            },
        )
        return HttpMongoCursor(documents)

    def find_one(
        self,
        filter: Optional[Dict[str, Any]] = None,
        projection: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Optional[Dict[str, Any]]:
        return self._rpc(
            "find_one",
            {
                "filter": filter or {},
                "projection": projection,
                "kwargs": kwargs,
            },
        )

    def aggregate(self, pipeline: Sequence[Dict[str, Any]], **kwargs: Any) -> Iterator[Dict[str, Any]]:
        documents = self._rpc(
            "aggregate",
            {"pipeline": list(pipeline), "kwargs": kwargs},
        )
        return iter(documents)

    def count_documents(self, filter: Optional[Dict[str, Any]] = None, **kwargs: Any) -> int:
        return self._rpc(
            "count_documents",
            {"filter": filter or {}, "kwargs": kwargs},
        )

    def insert_one(self, document: Dict[str, Any], **_: Any) -> InsertOneResult:
        return InsertOneResult(**self._rpc("insert_one", {"document": document}))

    def insert_many(self, documents: List[Dict[str, Any]], ordered: bool = True, **_: Any) -> InsertManyResult:
        return InsertManyResult(**self._rpc("insert_many", {"documents": documents, "ordered": ordered}))

    def update_one(self, filter: Dict[str, Any], update: Dict[str, Any], upsert: bool = False, **_: Any) -> UpdateResult:
        return UpdateResult(**self._rpc("update_one", {"filter": filter, "update": update, "upsert": upsert}))

    def update_many(self, filter: Dict[str, Any], update: Dict[str, Any], upsert: bool = False, **_: Any) -> UpdateResult:
        return UpdateResult(**self._rpc("update_many", {"filter": filter, "update": update, "upsert": upsert}))

    def delete_one(self, filter: Dict[str, Any], **_: Any) -> DeleteResult:
        return DeleteResult(**self._rpc("delete_one", {"filter": filter}))

    def delete_many(self, filter: Dict[str, Any], **_: Any) -> DeleteResult:
        return DeleteResult(**self._rpc("delete_many", {"filter": filter}))

    def drop(self, **_: Any) -> bool:
        return bool(self._rpc("drop", {}))

    def create_index(self, keys: List[Tuple[str, int]], **kwargs: Any) -> str:
        return self._rpc("create_index", {"keys": keys, "kwargs": kwargs})

    def index_information(self) -> Dict[str, Any]:
        return self._rpc("index_information", {})

    def drop_index(self, name: str, **_: Any) -> bool:
        return bool(self._rpc("drop_index", {"name": name}))



from __future__ import annotations

from mongoengine import Document


class HttpDocument(Document):
    """Common base class for models using the HTTP Mongo gateway."""

    meta = {"abstract": True}



"""Media I/O abstraction layer for :class:`~yggdrasil.io.buffer.BytesIO`.

Each concrete :class:`MediaIO` subclass handles reading and writing a specific
columnar or container format (Parquet, JSON, Arrow IPC, ZIP). The base class
provides:

* **Transparent compression** — when :attr:`media_type` carries a
  :class:`~yggdrasil.io.enums.Codec` (e.g. ``application/json+gzip``),
  reads automatically decompress and writes automatically compress, so
  callers never need to wrap in manual ``decompress()`` / ``compress()``
  calls.
* **Generic adapters** — :meth:`read_pylist`, :meth:`read_pydict`,
  :meth:`read_pandas_frame`, :meth:`read_polars_frame` (and their write
  counterparts) are all routed through the Arrow table read/write path.
* **Batched iteration** — every read method accepts an optional
  ``batch_size`` parameter. When set to a positive integer the method
  returns an ``Iterator`` of chunks instead of a single object, enabling
  memory-efficient streaming over large datasets.

Subclasses only need to implement three methods:

* ``check_options(options, **kwargs) → O`` — validate/merge caller options.
* ``_read_arrow_batches(options) → Iterator[pyarrow.RecordBatch]`` — yield
  record batches from the *uncompressed* buffer.
* ``_write_arrow_batches(batches, schema, options)`` — consume an iterator
  of record batches and write into the *uncompressed* buffer.

Typical usage::

    buf = BytesIO(raw_gzipped_parquet_bytes)
    buf.set_media_type(MediaType("application/vnd.apache.parquet+gzip"), safe=False)
    mio = buf.media_io()                     # returns ParquetIO with auto-decompress
    table = mio.read_arrow_table()           # decompresses then reads

    # Batched iteration
    for batch in mio.read_arrow_table(batch_size=1000):
        process(batch)
"""
from __future__ import annotations

import itertools
from abc import ABC, abstractmethod
from collections.abc import Iterable
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Generic, Iterator, Optional, Sequence, TypeVar, Union

import pyarrow as pa
from yggdrasil.arrow.cast import cast_arrow_tabular
from yggdrasil.io import MimeTypes
from yggdrasil.io.enums import Codec, MediaType, MimeType, SaveMode
from yggdrasil.pickle.serde import ObjectSerde

from .bytes_io import BytesIO
from .media_options import MediaOptions

if TYPE_CHECKING:
    import pandas
    import polars
    import pyarrow


O = TypeVar("O", bound=MediaOptions)

__all__ = ["MediaIO"]


@dataclass(slots=True)
class MediaIO(ABC, Generic[O]):
    """Abstract base for format-specific I/O on a :class:`BytesIO` buffer."""

    media_type: MediaType
    buffer: BytesIO

    # ------------------------------------------------------------------
    # Codec helpers (private)
    # ------------------------------------------------------------------

    @property
    def codec(self) -> Codec | None:
        """Return the transport compression codec, or ``None``."""
        return self.media_type.codec if self.media_type else None

    def _decompressed_buffer(self) -> tuple[BytesIO, bool]:
        """Return ``(buffer, was_decompressed)``.

        If the media type has a codec the buffer is decompressed into a
        **copy** (the original buffer is not mutated). The caller must
        close the copy when done.
        """
        codec = self.codec
        if codec is not None and self.buffer.size > 0:
            return self.buffer.decompress(codec=codec, copy=True), True
        return self.buffer, False

    def _compress_into_buffer(self, plain: BytesIO) -> None:
        """Compress *plain* with :attr:`codec` and replace :attr:`buffer` contents."""
        codec = self.codec
        if codec is not None:
            compressed = plain.compress(codec=codec, copy=True)
            self.buffer.replace_with_payload(compressed.to_bytes())
        else:
            self.buffer.replace_with_payload(plain.to_bytes())

    # ------------------------------------------------------------------
    # Cast helpers (private)
    # ------------------------------------------------------------------

    @staticmethod
    def _apply_cast(
        data: "pyarrow.Table | pyarrow.RecordBatch",
        *,
        options: O,
    ) -> "pyarrow.Table | pyarrow.RecordBatch":
        """Apply configured Arrow cast when requested."""
        cast = options.cast
        if cast is None:
            return data

        try:
            return cast_arrow_tabular(data, options=cast)
        except Exception:
            if options.raise_error:
                raise
            return data

    # ------------------------------------------------------------------
    # Iterable inference helpers (private)
    # ------------------------------------------------------------------

    @staticmethod
    def _is_nonstring_iterable(obj: Any) -> bool:
        """Return True for iterables not handled by dedicated branches."""
        return (
            isinstance(obj, Iterable)
            and not isinstance(obj, (str, bytes, bytearray, dict, list, pa.Table))
        )

    @staticmethod
    def _peek_iterable(obj: Iterable[Any]) -> tuple[Any, Iterator[Any]]:
        """Return first item and a rebuilt iterator including that first item."""
        iterator = iter(obj)
        first = next(iterator)
        return first, itertools.chain((first,), iterator)

    @classmethod
    def _table_from_tabular_iterable(cls, obj: Iterable[Any]) -> "pyarrow.Table":
        """Infer an Arrow table from an iterable of supported tabular objects."""
        first, iterator = cls._peek_iterable(obj)

        if isinstance(first, dict):
            return pa.Table.from_pylist(list(iterator)) # noqa

        if isinstance(first, pa.RecordBatch):
            return pa.Table.from_batches(list(iterator))

        if isinstance(first, pa.Table):
            tables = list(iterator)
            if not tables:
                return pa.table({})
            return pa.concat_tables(tables, promote_options="default")

        raise TypeError(
            "Unsupported iterable for write_table: expected an iterable of "
            "dict, pyarrow.RecordBatch, or pyarrow.Table, got first item "
            f"of type {type(first)!r}"
        )

    @classmethod
    def _iter_to_batches(
        cls,
        obj: Iterable[Any],
    ) -> tuple[Iterator["pyarrow.RecordBatch"], "pyarrow.Schema"]:
        """Infer a batch iterator and schema from an iterable of tabular objects."""
        first, iterator = cls._peek_iterable(obj)

        if isinstance(first, dict):
            table = pa.Table.from_pylist(list(iterator)) # noqa
            return iter(table.to_batches()), table.schema

        if isinstance(first, pa.RecordBatch):
            return iterator, first.schema

        if isinstance(first, pa.Table):
            def batch_iter() -> Iterator["pyarrow.RecordBatch"]:
                for tb in iterator:
                    yield from tb.to_batches()

            return batch_iter(), first.schema

        raise TypeError(
            "Unsupported iterable for write_table: expected an iterable of "
            "dict, pyarrow.RecordBatch, or pyarrow.Table, got first item "
            f"of type {type(first)!r}"
        )

    def _write_batches_iterable(
        self,
        obj: Iterable[Any],
        *,
        options: O,
    ) -> None:
        """Write an iterable of tabular chunks to the buffer."""
        try:
            batches, schema = self._iter_to_batches(obj)
        except StopIteration:
            self._write_single_table(pa.table({}), options)
            return

        codec = self.codec
        if codec is not None:
            plain = BytesIO(config=self.buffer.config)
            orig_buffer = self.buffer
            self.buffer = plain
            try:
                self._write_arrow_batches(
                    batches=batches,
                    schema=schema,
                    options=options,
                )
            finally:
                self.buffer = orig_buffer
            self._compress_into_buffer(plain)
            plain.close()
        else:
            self._write_arrow_batches(
                batches=batches,
                schema=schema,
                options=options,
            )

    # ------------------------------------------------------------------
    # Abstract protocol
    # ------------------------------------------------------------------

    @classmethod
    @abstractmethod
    def check_options(cls, options: Optional[O], *args, **kwargs) -> O:
        """Validate and merge caller-supplied options into a concrete instance."""

    def read_arrow_batches(
        self,
        *args,
        options: Optional[O] = None,
        **media_options,
    ) -> Iterator["pyarrow.RecordBatch"]:
        """Yield Arrow record batches, transparently handling decompression."""
        resolved = self.check_options(
            options=options, *args, **media_options
        )

        buf, decompressed = self._decompressed_buffer()
        orig_buffer = self.buffer
        try:
            if decompressed:
                self.buffer = buf

            for batch in self._read_arrow_batches(options=resolved):
                casted = self._apply_cast(batch, options=resolved)
                if isinstance(casted, pa.Table):
                    yield from casted.to_batches()
                else:
                    yield casted
        finally:
            if decompressed:
                self.buffer = orig_buffer
                buf.close()

    @abstractmethod
    def _read_arrow_batches(self, *, options: O) -> Iterator["pyarrow.RecordBatch"]:
        """Yield record batches from the **uncompressed** buffer."""

    def read_polars_frames(
        self,
        *args,
        options: Optional[O] = None,
        **media_options,
    ) -> Iterator["polars.DataFrame | polars.LazyFrame"]:
        from yggdrasil.polars.lib import polars as _pl

        resolved = self.check_options(
            options=options, *args, **media_options
        )
        for batch in self.read_arrow_batches(options=resolved):
            df = _pl.from_arrow(batch, rechunk=False)
            yield df.lazy() if resolved.lazy else df

    @abstractmethod
    def _write_arrow_batches(
        self,
        *,
        batches: Iterator["pyarrow.RecordBatch"],
        schema: "pyarrow.Schema",
        options: O,
    ) -> None:
        """Consume record batches and write into the **uncompressed** buffer."""

    # ------------------------------------------------------------------
    # Internal helpers: table ↔ batches bridge
    # ------------------------------------------------------------------

    def _read_table_from_batches(self, *, options: O) -> "pyarrow.Table":
        """Collect all batches from :meth:`read_arrow_batches` into one table."""
        batches = list(self.read_arrow_batches(options=options))
        if not batches:
            return pa.table({})
        return pa.Table.from_batches(batches)

    def _write_table_as_batches(self, *, table: "pyarrow.Table", options: O) -> None:
        """Convert *table* to batches and forward to :meth:`_write_arrow_batches`."""
        casted = options.cast.cast_arrow(table)

        self._write_arrow_batches(
            batches=iter(casted.to_batches()),
            schema=casted.schema,
            options=options.with_cast(None),
        )

    # ------------------------------------------------------------------
    # Factory
    # ------------------------------------------------------------------

    @classmethod
    def make(
        cls,
        buffer: BytesIO,
        media: MediaType | MimeType | str,
    ) -> "MediaIO[MediaOptions]":
        """Create the appropriate :class:`MediaIO` subclass for *media*."""
        media = MediaType.parse(media)
        mt = media.mime_type
        buffer.set_media_type(media, safe=False)

        if mt is MimeTypes.PARQUET:
            from .parquet_io import ParquetIO
            return ParquetIO(media_type=media, buffer=buffer)
        if mt is MimeTypes.JSON or mt is MimeTypes.NDJSON:
            from .json_io import JsonIO
            return JsonIO(media_type=media, buffer=buffer)
        if mt is MimeTypes.ZIP:
            from .zip_io import ZipIO
            return ZipIO(media_type=media, buffer=buffer)
        if mt is MimeTypes.ARROW_IPC:
            from .arrow_ipc_io import IPCIO
            return IPCIO(media_type=media, buffer=buffer)

        raise NotImplementedError(f"Cannot create media IO for {media!r}")

    # ------------------------------------------------------------------
    # Write guard
    # ------------------------------------------------------------------

    def skip_write(self, mode: SaveMode) -> bool:
        """Return ``True`` when to write should be skipped."""
        if self.buffer.size > 0:
            if mode == SaveMode.IGNORE:
                return True
            if mode == SaveMode.ERROR_IF_EXISTS:
                raise IOError(
                    f"Cannot write in already existing {self.buffer!r} "
                    f"with save mode {SaveMode.ERROR_IF_EXISTS.value}"
                )
        return False

    # ------------------------------------------------------------------
    # Generic write dispatcher
    # ------------------------------------------------------------------

    def write_table(
        self,
        obj: Any,
        *,
        options: O | None = None,
        **option_kwargs,
    ) -> None:
        """Write a supported tabular object to the buffer.

        Supported inputs
        ----------------
        - pyarrow.Table
        - Iterator[pyarrow.Table]
        - Iterator[pyarrow.RecordBatch]
        - pandas.DataFrame
        - polars.DataFrame
        - polars.LazyFrame
        - list[dict]
        - Iterator[dict]
        - dict[str, list]
        """
        if isinstance(obj, pa.Table):
            return self.write_arrow_table(
                table=obj,
                options=options,
                **option_kwargs,
            )
        elif isinstance(obj, pa.RecordBatch):
            return self.write_arrow_table(
                pa.Table.from_batches([obj]),
                options=options,
                **option_kwargs,
            )

        ns, _ = ObjectSerde.module_and_name(obj)

        if ns.startswith("polars"):
            return self.write_polars_frame(
                obj,
                options=options,
                **option_kwargs,
            )
        elif ns.startswith("pandas"):
            return self.write_pandas_frame(
                obj,
                options=options,
                **option_kwargs
            )

        if isinstance(obj, list):
            if obj and not all(isinstance(row, dict) for row in obj):
                raise TypeError(
                    "write_table(list) expects list[dict], "
                    f"got list[{type(obj[0]).__name__}]"
                )
            self.write_pylist(
                data=obj,
                options=options,
                **option_kwargs,
            )
            return

        if isinstance(obj, dict):
            self.write_pydict(
                data=obj,
                options=options,
                **option_kwargs,
            )
            return

        if self._is_nonstring_iterable(obj):
            resolved = self.check_options(options=options, **option_kwargs)
            if self.skip_write(mode=resolved.mode):
                return

            if resolved.mode in (SaveMode.APPEND, SaveMode.UPSERT):
                table = self._table_from_tabular_iterable(obj)
                existing = self._read_existing_table()
                table = self._apply_save_mode(
                    existing,
                    table,
                    mode=resolved.mode,
                    match_by=resolved.match_by,
                )
                self._write_single_table(table, resolved)
                return

            self._write_batches_iterable(obj, options=resolved)
            return

        ns = ObjectSerde.full_namespace(obj)

        if ns.startswith("pandas."):
            self.write_pandas_frame(
                frame=obj,
                options=options,
                **option_kwargs,
            )
            return

        if ns.startswith("polars."):
            self.write_polars_frame(
                frame=obj,
                options=options,
                **option_kwargs,
            )
            return

        raise TypeError(
            "Unsupported tabular object for write_table: "
            f"{type(obj)!r} (namespace={ns!r}). Expected one of: "
            "pyarrow.Table, Iterator[pyarrow.Table], Iterator[pyarrow.RecordBatch], "
            "pandas.DataFrame, polars.DataFrame, polars.LazyFrame, "
            "list[dict], Iterator[dict], dict[str, list]."
        )

    # ------------------------------------------------------------------
    # Save-mode helpers (private)
    # ------------------------------------------------------------------

    def _read_existing_table(self) -> "pyarrow.Table | None":
        """Read the current buffer contents as an Arrow table, or ``None``."""
        if self.buffer.size <= 0:
            return None
        try:
            table = self.read_arrow_table()
            return table if isinstance(table, pa.Table) else None
        except Exception:
            return None

    @staticmethod
    def _apply_save_mode(
        existing: "pyarrow.Table | None",
        incoming: "pyarrow.Table",
        *,
        mode: SaveMode,
        match_by: "Sequence[str] | None",
    ) -> "pyarrow.Table":
        """Combine *existing* and *incoming* tables according to *mode*."""
        if existing is None or existing.num_rows == 0:
            return incoming

        if mode in (SaveMode.AUTO, SaveMode.OVERWRITE, SaveMode.TRUNCATE):
            return incoming

        if mode == SaveMode.APPEND:
            return pa.concat_tables([existing, incoming], promote_options="default")

        if mode == SaveMode.UPSERT:
            if not match_by:
                raise ValueError(
                    "SaveMode.UPSERT requires match_by columns, got None/empty"
                )
            match_by = list(match_by)

            for col in match_by:
                if col not in existing.column_names:
                    raise ValueError(
                        f"UPSERT match_by column {col!r} not in existing table "
                        f"(columns: {existing.column_names})"
                    )
                if col not in incoming.column_names:
                    raise ValueError(
                        f"UPSERT match_by column {col!r} not in incoming table "
                        f"(columns: {incoming.column_names})"
                    )

            incoming_keys: set[tuple] = set()
            incoming_key_arrays = [
                incoming.column(c).to_pylist() for c in match_by
            ]
            for row_idx in range(incoming.num_rows):
                key = tuple(arr[row_idx] for arr in incoming_key_arrays)
                incoming_keys.add(key)

            existing_key_arrays = [
                existing.column(c).to_pylist() for c in match_by
            ]
            keep_mask: list[bool] = []
            for row_idx in range(existing.num_rows):
                key = tuple(arr[row_idx] for arr in existing_key_arrays)
                keep_mask.append(key not in incoming_keys)

            surviving = existing.filter(pa.array(keep_mask, type=pa.bool_()))

            return pa.concat_tables(
                [surviving, incoming],
                promote_options="default",
            )

        return incoming

    # ------------------------------------------------------------------
    # Batching helpers (private)
    # ------------------------------------------------------------------

    @staticmethod
    def _iter_arrow_batches(
        table: "pyarrow.Table",
        batch_size: int,
    ) -> Iterator["pyarrow.Table"]:
        """Yield successive :class:`pyarrow.Table` slices of *batch_size* rows."""
        total = table.num_rows
        for offset in range(0, total, batch_size):
            yield table.slice(offset, min(batch_size, total - offset))

    # ------------------------------------------------------------------
    # Arrow (primary public API)
    # ------------------------------------------------------------------

    def read_arrow_table(
        self,
        *,
        options: O | None = None,
        **option_kwargs,
    ) -> Union["pyarrow.Table", Iterator["pyarrow.Table"]]:
        """Read the buffer into a :class:`pyarrow.Table`."""
        resolved = self.check_options(options=options, **option_kwargs)
        table = self._read_table_from_batches(options=resolved)

        bs = resolved.batch_size
        if bs and bs > 0:
            return self._iter_arrow_batches(table, bs)
        return table

    def write_arrow_table(
        self,
        table: "pyarrow.Table",
        *,
        options: O | None = None,
        **option_kwargs,
    ) -> None:
        """Write a :class:`pyarrow.Table` to the buffer."""
        resolved = self.check_options(
            options=options, **option_kwargs,
        )
        if self.skip_write(mode=resolved.mode):
            return

        if resolved.mode in (SaveMode.APPEND, SaveMode.UPSERT):
            existing = self._read_existing_table()
            table = self._apply_save_mode(
                existing,
                table,
                mode=resolved.mode,
                match_by=resolved.match_by,
            )

        self._write_single_table(table, resolved)

    def _write_single_table(
        self,
        table: "pyarrow.Table",
        options: O,
    ) -> None:
        """Write one table handling codec compression."""
        codec = self.codec
        if codec is not None:
            plain = BytesIO(config=self.buffer.config)
            orig_buffer = self.buffer
            self.buffer = plain
            try:
                self._write_table_as_batches(table=table, options=options)
            finally:
                self.buffer = orig_buffer
            self._compress_into_buffer(plain)
            plain.close()
        else:
            self._write_table_as_batches(table=table, options=options)

    # ------------------------------------------------------------------
    # Python dict / list convenience wrappers
    # ------------------------------------------------------------------

    def read_pylist(
        self,
        *,
        options: O | None = None,
        **option_kwargs,
    ) -> Union[list[dict], Iterator[list[dict]]]:
        """Read the buffer as a list of row dicts."""
        resolved = self.check_options(options=options, **option_kwargs)
        bs = resolved.batch_size
        if bs and bs > 0:
            return (
                chunk.to_pylist()
                for chunk in self.read_arrow_table(options=resolved)
            )
        return self.read_arrow_table(options=resolved).to_pylist()

    def write_pylist(
        self,
        data: list[dict],
        *,
        options: O | None = None,
        **option_kwargs,
    ) -> None:
        """Write a list of row dicts to the buffer."""
        tb = pa.Table.from_pylist(data) # noqa
        self.write_arrow_table(
            table=tb,
            options=options,
            **option_kwargs,
        )

    def read_pydict(
        self,
        *,
        options: O | None = None,
        **option_kwargs,
    ) -> Union[dict[str, list], Iterator[dict[str, list]]]:
        """Read the buffer as a column-oriented dict."""
        resolved = self.check_options(options=options, **option_kwargs)
        bs = resolved.batch_size
        if bs and bs > 0:
            return (
                chunk.to_pydict()
                for chunk in self.read_arrow_table(options=resolved)
            )
        return self.read_arrow_table(options=resolved).to_pydict()

    def write_pydict(
        self,
        data: dict[str, list],
        *,
        options: O | None = None,
        **option_kwargs,
    ) -> None:
        """Write a column-oriented dict to the buffer."""
        tb = pa.Table.from_pydict(data) # noqa
        self.write_arrow_table(
            table=tb,
            options=options,
            **option_kwargs,
        )

    # ------------------------------------------------------------------
    # Pandas / Polars convenience wrappers
    # ------------------------------------------------------------------

    def read_pandas_frame(
        self,
        *,
        options: O | None = None,
        **option_kwargs,
    ) -> Union["pandas.DataFrame", Iterator["pandas.DataFrame"]]:
        """Read the buffer as a :class:`pandas.DataFrame`."""
        resolved = self.check_options(options=options, **option_kwargs)
        bs = resolved.batch_size
        if bs and bs > 0:
            return (
                chunk.to_pandas()
                for chunk in self.read_arrow_table(options=resolved)
            )
        return self.read_arrow_table(options=resolved).to_pandas()

    def write_pandas_frame(
        self,
        frame: "pandas.DataFrame",
        *,
        options: O | None = None,
        **option_kwargs,
    ) -> None:
        """Write a :class:`pandas.DataFrame` to the buffer."""
        tb = pa.Table.from_pandas(
            frame,
            preserve_index=bool(frame.index.name),
        ) # noqa

        self.write_arrow_table(
            table=tb,
            options=options,
            **option_kwargs,
        )

    def read_polars_frame(
        self,
        *,
        options: O | None = None,
        **option_kwargs,
    ) -> Union[
        "polars.DataFrame",
        "polars.LazyFrame",
        Iterator["polars.DataFrame"],
    ]:
        """Read the buffer as a Polars frame."""
        from yggdrasil.polars.lib import polars as _pl

        resolved = self.check_options(options=options, **option_kwargs)
        bs = resolved.batch_size
        if bs and bs > 0:
            return (
                _pl.from_arrow(chunk, rechunk=False)
                for chunk in self.read_arrow_table(options=resolved)
            )

        tb = self.read_arrow_table(options=resolved)
        df = _pl.from_arrow(tb, rechunk=False)
        return df.lazy() if resolved.lazy else df

    def write_polars_frame(
        self,
        frame: "polars.DataFrame | polars.LazyFrame",
        *,
        options: O | None = None,
        **option_kwargs,
    ) -> None:
        """Write a Polars DataFrame or LazyFrame to the buffer."""
        from yggdrasil.polars.cast import polars_dataframe_to_arrow_table

        resolved = self.check_options(
            options=options, **option_kwargs,
        )

        tb = polars_dataframe_to_arrow_table(frame, resolved.cast)

        self.write_arrow_table(
            table=tb,
            options=options,
            **option_kwargs,
        )


from __future__ import annotations

from dataclasses import dataclass
from typing import Any, List, Optional


@dataclass(slots=True)
class InsertOneResult:
    inserted_id: Any
    acknowledged: bool = True


@dataclass(slots=True)
class InsertManyResult:
    inserted_ids: List[Any]
    acknowledged: bool = True


@dataclass(slots=True)
class UpdateResult:
    matched_count: int
    modified_count: int
    upserted_id: Optional[Any] = None
    acknowledged: bool = True


@dataclass(slots=True)
class DeleteResult:
    deleted_count: int
    acknowledged: bool = True



from __future__ import annotations

from typing import Any, Dict, Iterable, Iterator, List


class HttpMongoCursor:
    """Minimal eager cursor wrapper good enough for MongoEngine aggregation use."""

    def __init__(self, documents: Iterable[Dict[str, Any]]) -> None:
        self._documents: List[Dict[str, Any]] = list(documents)

    def __iter__(self) -> Iterator[Dict[str, Any]]:
        return iter(self._documents)

    def __len__(self) -> int:
        return len(self._documents)

    def to_list(self) -> List[Dict[str, Any]]:
        return list(self._documents)



from __future__ import annotations

from typing import Any, Dict, List

from .collection import HttpCollection
from .transport import HttpTransport


class HttpDatabase:
    def __init__(self, transport: HttpTransport, name: str) -> None:
        self._transport = transport
        self.name = name

    def __getitem__(self, collection_name: str) -> HttpCollection:
        return HttpCollection(self._transport, self.name, collection_name)

    def get_collection(self, name: str, **_: Any) -> HttpCollection:
        return self[name]

    def list_collection_names(self) -> List[str]:
        return self._transport.rpc(database=self.name, operation="list_collection_names", arguments={})

    def command(self, command: Any, **kwargs: Any) -> Dict[str, Any]:
        if isinstance(command, str):
            payload = {command: 1, **kwargs} if kwargs else {command: 1}
        else:
            payload = command
        return self._transport.rpc(database=self.name, operation="command", arguments={"command": payload})



"""yggdrasil.concurrent — async job primitives and bounded thread-pool."""

from .job import Job
from .job_result import JobResult
from .pool import JobPoolExecutor
# Thread-specific classes live in .threading (their canonical module)
from .threading import AsyncJob, ThreadJob

__all__ = ["AsyncJob", "Job", "ThreadJob", "JobResult", "JobPoolExecutor"]



from __future__ import annotations

import os
from typing import Any, Dict, Optional

import requests

from .exceptions import HttpMongoError, HttpMongoTransportError
from .serde import BsonJsonSerde


class HttpTransport:
    """Tiny RPC transport over HTTP for the Mongo gateway."""

    def __init__(
        self,
        host: Optional[str] = None,
        token: Optional[str] = None,
        timeout: float = 30.0,
        session: Optional[requests.Session] = None,
        **_: Any,
    ) -> None:
        self.base_url = (host or os.environ.get("MONGODB_URI") or "http://127.0.0.1:8000").rstrip("/")
        self.timeout = timeout
        self.session = session or requests.Session()
        self.serde = BsonJsonSerde()
        self.headers = {
            "Content-Type": "application/json",
            "X-Auth-Token": token or os.environ.get("HTTPMONGO_TOKEN", "dev-secret"),
        }

    def close(self) -> None:
        self.session.close()

    def rpc(
        self,
        *,
        database: str,
        operation: str,
        arguments: Dict[str, Any],
        collection: Optional[str] = None,
    ) -> Any:
        payload = {
            "database": database,
            "collection": collection,
            "operation": operation,
            "arguments": arguments,
        }

        try:
            response = self.session.post(
                f"{self.base_url}/rpc",
                headers=self.headers,
                data=self.serde.dumps(payload),
                timeout=self.timeout,
            )
        except requests.RequestException as exc:
            raise HttpMongoTransportError(f"request failed: {exc}") from exc

        try:
            data = self.serde.loads(response.text)
        except Exception as exc:  # noqa: BLE001
            raise HttpMongoTransportError(
                f"invalid response: status={response.status_code}, body={response.text[:500]}"
            ) from exc

        if not response.ok:
            raise HttpMongoTransportError(f"http error {response.status_code}: {data}")

        if data.get("ok") != 1:
            error = data.get("error", {})
            raise HttpMongoError(f"{error.get('type', 'RemoteError')}: {error.get('message', data)}")

        return data.get("result")



from __future__ import annotations

import json
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

import pyarrow as pa

if TYPE_CHECKING:
    import polars as pl
    import pyspark.sql.types as pst


__all__ = [
    "Field",
    "field",
    "_normalize_metadata",
    "_to_bytes",
    "_merge_metadata_and_tags",
    "_decode_metadata_dict",
]


def _to_bytes(value: object) -> bytes:
    if value is None:
        return b""
    if isinstance(value, bytes):
        return value
    if isinstance(value, str):
        return value.encode("utf-8")
    return json.dumps(value, separators=(",", ":"), ensure_ascii=False).encode("utf-8")


def _normalize_metadata(
    metadata: dict[bytes | str, bytes | str | object] | None,
    tags: dict[bytes | str, bytes | str | object] | None,
) -> dict[bytes, bytes] | None:
    if not metadata and not tags:
        return None

    normalized = {
        _to_bytes(key): _to_bytes(value)
        for key, value in (metadata or {}).items()
        if key and value is not None
    }

    if tags:
        normalized.update(
            {
                b"t:" + _to_bytes(key): _to_bytes(value)
                for key, value in tags.items()
                if key and value is not None
            }
        )

    return normalized or None


def _merge_metadata_and_tags(
    metadata: dict[bytes, bytes] | None,
    tags: dict[bytes, bytes] | None,
) -> dict[bytes, bytes] | None:
    merged: dict[bytes, bytes] = dict(metadata or {})

    if tags:
        merged.update(
            {
                key if key.startswith(b"t:") else b"t:" + key: value
                for key, value in tags.items()
            }
        )

    return merged or None


def _decode_metadata_dict(
    metadata: dict[bytes, bytes] | None,
) -> dict[str, object]:
    if not metadata:
        return {}

    result: dict[str, object] = {}
    for key, value in metadata.items():
        decoded_key = key.decode("utf-8") if isinstance(key, bytes) else str(key)

        try:
            result[decoded_key] = json.loads(value.decode("utf-8"))
        except Exception:
            result[decoded_key] = value.decode("utf-8", errors="replace")

    return result


def field(
    name: str,
    arrow_type: pa.DataType,
    *,
    nullable: bool = True,
    metadata: dict[bytes | str, bytes | str | object] | None = None,
    tags: dict[bytes | str, bytes | str | object] | None = None,
) -> "Field":
    return Field(
        name=name,
        arrow_type=arrow_type,
        nullable=nullable,
        metadata=_normalize_metadata(metadata, tags=tags),
    )


@dataclass(frozen=True, slots=True)
class Field:
    name: str
    arrow_type: pa.DataType
    nullable: bool = True
    metadata: dict[bytes, bytes] | None = None

    @property
    def partition_by(self) -> bool:
        if self.metadata:
            v = self.metadata.get(b"t:partition_by")
            if v:
                return v.startswith(b"t")
        return False

    @property
    def cluster_by(self) -> bool:
        if self.metadata:
            v = self.metadata.get(b"t:cluster_by")
            if v:
                return v.startswith(b"t")
        return False

    @property
    def tags(self) -> dict[bytes, bytes]:
        if not self.metadata:
            return {}

        return {k[2:]: v for k, v in self.metadata.items() if k.startswith(b"t:")}

    def copy(
        self,
        *,
        name: str | None = None,
        arrow_type: pa.DataType | None = None,
        nullable: bool | None = None,
        metadata: dict[bytes | str, bytes | str | object] | None = None,
        tags: dict[bytes | str, bytes | str | object] | None = None,
    ) -> "Field":
        return Field(
            name=self.name if name is None else name,
            arrow_type=self.arrow_type if arrow_type is None else arrow_type,
            nullable=self.nullable if nullable is None else nullable,
            metadata=(
                (dict(self.metadata) if self.metadata is not None else None)
                if metadata is None and tags is None
                else _normalize_metadata(metadata, tags=tags)
            ),
        )

    def autotag(self) -> "Field":
        inferred: dict[str, object] = {}

        name = self.name.lower()
        dtype = self.arrow_type

        # ---- dtype family tags
        if pa.types.is_boolean(dtype):
            inferred["kind"] = "boolean"
        elif pa.types.is_integer(dtype):
            inferred["kind"] = "integer"
            inferred["numeric"] = True
        elif pa.types.is_floating(dtype):
            inferred["kind"] = "float"
            inferred["numeric"] = True
        elif pa.types.is_decimal(dtype):
            inferred["kind"] = "decimal"
            inferred["numeric"] = True
        elif pa.types.is_timestamp(dtype):
            inferred["kind"] = "timestamp"
            inferred["temporal"] = True
        elif pa.types.is_date(dtype):
            inferred["kind"] = "date"
            inferred["temporal"] = True
        elif pa.types.is_time(dtype):
            inferred["kind"] = "time"
            inferred["temporal"] = True
        elif pa.types.is_duration(dtype):
            inferred["kind"] = "duration"
            inferred["temporal"] = True
        elif pa.types.is_string(dtype) or pa.types.is_large_string(dtype):
            inferred["kind"] = "string"
        elif pa.types.is_binary(dtype) or pa.types.is_large_binary(dtype):
            inferred["kind"] = "binary"
        elif (
            pa.types.is_list(dtype)
            or pa.types.is_large_list(dtype)
            or pa.types.is_fixed_size_list(dtype)
        ):
            inferred["kind"] = "list"
            inferred["nested"] = True
        elif pa.types.is_struct(dtype):
            inferred["kind"] = "struct"
            inferred["nested"] = True
        elif pa.types.is_map(dtype):
            inferred["kind"] = "map"
            inferred["nested"] = True
        else:
            inferred["kind"] = str(dtype)

        inferred["nullable"] = self.nullable

        # ---- timezone from Arrow timestamp type
        if pa.types.is_timestamp(dtype):
            unit = getattr(dtype, "unit", None)

            if unit:
                inferred["unit"] = str(unit)

            tz = getattr(dtype, "tz", None)
            if tz:
                inferred["tz"] = tz

        # ---- semantic name tags
        if name == "id" or name.endswith("_id") or name.endswith("id"):
            inferred.setdefault("role", "identifier")

        if (
            name in {"ts", "timestamp"}
            or name.endswith("_ts")
            or name.endswith("_timestamp")
        ):
            inferred["temporal"] = True
            inferred.setdefault("role", "event_time")

        if name == "date" or name.endswith("_date"):
            inferred["temporal"] = True
            inferred.setdefault("role", "date")

        if name == "dt" or name.endswith("_dt"):
            inferred["temporal"] = True

        if "created_at" in name:
            inferred["temporal"] = True
            inferred["role"] = "created_at"

        if "updated_at" in name:
            inferred["temporal"] = True
            inferred["role"] = "updated_at"

        if "deleted_at" in name:
            inferred["temporal"] = True
            inferred["role"] = "deleted_at"

        if name.startswith("is_") or name.startswith("has_") or name.endswith("_flag"):
            inferred.setdefault("role", "flag")

        if "price" in name:
            inferred.setdefault("role", "price")
            inferred["numeric"] = True

        if any(
            token in name
            for token in ("qty", "quantity", "volume", "count", "size", "amount")
        ):
            inferred.setdefault("role", "measure")
            inferred["numeric"] = True

        if any(
            token in name
            for token in ("name", "label", "desc", "description", "comment")
        ):
            inferred.setdefault("role", "attribute")

        if any(
            token in name
            for token in ("country", "region", "area", "zone", "market", "book", "desk")
        ):
            inferred.setdefault("role", "dimension")

        if self.partition_by:
            inferred["partition_by"] = True

        if self.cluster_by:
            inferred["cluster_by"] = True

        # explicit existing tags win
        merged_tags: dict[bytes, bytes] = {
            _to_bytes(key): _to_bytes(value)
            for key, value in inferred.items()
            if key and value is not None
        }
        merged_tags.update(self.tags)

        return Field(
            name=self.name,
            arrow_type=self.arrow_type,
            nullable=self.nullable,
            metadata=_merge_metadata_and_tags(self.metadata, merged_tags),
        )

    @classmethod
    def from_any(cls, obj: Any) -> "Field":
        if isinstance(obj, cls):
            return obj

        from yggdrasil.arrow.cast import any_to_arrow_field

        return cls.from_arrow(any_to_arrow_field(obj))

    @classmethod
    def from_arrow(cls, value: pa.Field) -> "Field":
        if not isinstance(value, pa.Field):
            from yggdrasil.arrow.cast import any_to_arrow_field
            value = any_to_arrow_field(value)

        return cls(
            name=value.name,
            arrow_type=value.type,
            nullable=value.nullable,
            metadata=value.metadata,
        )

    def to_arrow_field(self) -> pa.Field:
        return pa.field(
            name=self.name,
            type=self.arrow_type,
            nullable=self.nullable,
            metadata=self.metadata,
        )

    @classmethod
    def from_polars(
        cls,
        obj: Any = None,
        *,
        name: str | None = None,
        dtype: "pl.DataType | None" = None,
        nullable: bool = True,
        metadata: dict[bytes | str, bytes | str | object] | None = None,
        tags: dict[bytes | str, bytes | str | object] | None = None,
    ) -> "Field":
        if obj is not None:
            from yggdrasil.arrow.cast import any_to_arrow_field
            return cls.from_arrow(any_to_arrow_field(obj))

        if name is None or dtype is None:
            raise ValueError("name and dtype are required when obj is not provided")

        from yggdrasil.polars.cast import polars_type_to_arrow_type

        return cls(
            name=name,
            arrow_type=polars_type_to_arrow_type(dtype),
            nullable=nullable,
            metadata=_normalize_metadata(metadata, tags=tags),
        )

    def to_polars_field(self) -> "pl.Field":
        from yggdrasil.polars.cast import arrow_field_to_polars_field

        return arrow_field_to_polars_field(self.to_arrow_field())

    @classmethod
    def from_pyspark(
        cls,
        obj: Any = None,
        *,
        name: str | None = None,
        dtype: "pst.DataType | None" = None,
        nullable: bool = True,
        metadata: dict[bytes | str, bytes | str | object] | None = None,
        tags: dict[bytes | str, bytes | str | object] | None = None,
    ) -> "Field":
        if obj is not None:
            from pyspark.sql.types import StructField
            from yggdrasil.spark.cast import spark_field_to_arrow_field

            if isinstance(obj, StructField):
                return cls.from_arrow(spark_field_to_arrow_field(obj))

            from yggdrasil.arrow.cast import any_to_arrow_field
            return cls.from_arrow(any_to_arrow_field(obj))

        if name is None or dtype is None:
            raise ValueError("name and dtype are required when obj is not provided")

        from yggdrasil.spark.cast import spark_type_to_arrow_type

        return cls(
            name=name,
            arrow_type=spark_type_to_arrow_type(dtype),
            nullable=nullable,
            metadata=_normalize_metadata(metadata, tags=tags),
        )

    def to_pyspark_field(self) -> "pst.StructField":
        from yggdrasil.spark.cast import arrow_field_to_spark_field

        return arrow_field_to_spark_field(self.to_arrow_field())



"""File-like IO abstractions for Databricks paths.

Each subclass of :class:`DatabricksIO` implements ``read_byte_range`` and
``write_all_bytes`` for exactly one Databricks namespace (DBFS, Workspace,
Volumes).  The base class provides the full buffered read/write/seek/flush
layer on top.

Buffers use :class:`yggdrasil.io.BytesIO` (``_Buffer``) which spills to
disk transparently for large payloads.
"""
from __future__ import annotations

import base64
import io
import logging
import time
from abc import ABC, abstractmethod
from threading import Thread
from typing import TYPE_CHECKING, Optional, IO, AnyStr, Union, BinaryIO

from yggdrasil.io.buffer import BytesIO as _Buffer

from databricks.sdk.errors import InternalError
from databricks.sdk.errors.platform import (
    NotFound,
    ResourceDoesNotExist,
    BadRequest,
)
from databricks.sdk.service.workspace import ImportFormat, ExportFormat

from .path_kind import DatabricksPathKind
from ...pyutils.retry import retry

if TYPE_CHECKING:
    from .path import DatabricksPath

__all__ = ["DatabricksIO"]

LOGGER = logging.getLogger(__name__)
_COPY_CHUNK = 8 * 1024 * 1024


# ---------------------------------------------------------------------------
# Helper: normalise arbitrary data into (_Buffer, size, owns_buffer)
# ---------------------------------------------------------------------------

def _prepare_buffer(
    data: Union[bytes, bytearray, memoryview, BinaryIO],
) -> tuple[int, _Buffer, bool]:
    """Wrap *data* in a :class:`_Buffer` and return ``(size, buf, True)``."""
    buf = _Buffer(data)
    return buf.size, buf, True


# ═══════════════════════════════════════════════════════════════════════════
# Base class
# ═══════════════════════════════════════════════════════════════════════════

class DatabricksIO(ABC, IO):
    """Buffered file-like interface for Databricks workspace/volume/DBFS paths.

    The buffer is a :class:`_Buffer` (``yggdrasil.io.BytesIO``) which spills
    transparently to disk for large payloads — no OOM risk.
    """

    # ── Construction ──────────────────────────────────────────────────

    def __init__(
        self,
        path: "DatabricksPath",
        mode: str,
        encoding: Optional[str] = None,
        position: int = 0,
        buffer: Optional[_Buffer] = None,
    ):
        super().__init__()
        self.encoding = encoding
        self.mode = mode          # triggers the property setter below
        self.path = path
        self._buffer: Optional[_Buffer] = buffer
        self.position = position
        self._write_flag = False

    # ── Factory: dispatch on path.kind ────────────────────────────────

    @classmethod
    def create_instance(
        cls,
        path: "DatabricksPath",
        mode: str,
        encoding: Optional[str] = None,
        position: int = 0,
        buffer: Optional[_Buffer] = None,
    ) -> "DatabricksIO":
        """Return the right IO subclass for *path.kind*."""
        _map = {
            DatabricksPathKind.VOLUME:    DatabricksVolumeIO,
            DatabricksPathKind.DBFS:      DatabricksDBFSIO,
            DatabricksPathKind.WORKSPACE: DatabricksWorkspaceIO,
        }
        klass = _map.get(path.kind)
        if klass is None:
            raise ValueError(f"Unsupported path kind: {path.kind}")
        return klass(
            path=path, mode=mode, encoding=encoding,
            position=position, buffer=buffer,
        )

    # ── Context manager ───────────────────────────────────────────────

    def __enter__(self) -> "DatabricksIO":
        return self.connect(clone=False)

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def __del__(self):
        if self._need_flush():
            try:
                Thread(target=self.close).start()
            except BaseException:
                pass

    # ── Dunder helpers ────────────────────────────────────────────────

    def __next__(self):
        line = self.readline()
        if not line:
            raise StopIteration
        return line

    def __len__(self):
        return self.content_length or 0

    def __iter__(self):
        return iter(self.read_all_bytes())

    def __hash__(self):
        return hash(self.path)

    def __str__(self):
        return f"{self.__class__.__name__}(path={self.path!r})"

    __repr__ = __str__

    # ── Properties ────────────────────────────────────────────────────

    @property
    def workspace(self):
        return self.path.workspace

    @property
    def name(self):
        return self.path.name

    @property
    def mode(self) -> str:
        return self._mode

    @mode.setter
    def mode(self, value: str):
        self._mode = value
        if "b" in value:
            self.encoding = None
        elif self.encoding is None:
            self.encoding = "utf-8"

    @property
    def content_length(self) -> int:
        return self.path.content_length

    @content_length.setter
    def content_length(self, value: int):
        self.path.content_length = value

    # ── Buffer management ─────────────────────────────────────────────

    @property
    def buffer(self) -> _Buffer:
        """Lazily create or return the spill-to-disk buffer."""
        if self._buffer is None:
            self._buffer = _Buffer()
            self._buffer.seek(self.position, io.SEEK_SET)
        return self._buffer

    @buffer.setter
    def buffer(self, value: Optional[_Buffer]) -> None:
        self._buffer = value

    def clear_buffer(self):
        self._buffer = None

    # ── Clone / connect ───────────────────────────────────────────────

    def clone_instance(self, **kwargs) -> "DatabricksIO":
        return self.__class__(
            path=kwargs.get("path", self.path),
            mode=kwargs.get("mode", self.mode),
            encoding=kwargs.get("encoding", self.encoding),
            position=kwargs.get("position", self.position),
            buffer=kwargs.get("buffer", self._buffer),
        )

    @property
    def connected(self) -> bool:
        return self.path.connected

    def connect(self, clone: bool = False) -> "DatabricksIO":
        path = self.path.connect()
        if clone:
            return self.clone_instance(path=path)
        self.path = path
        return self

    # ── Lifecycle ─────────────────────────────────────────────────────

    def close(self, flush: bool = True):
        if flush:
            self.flush()
        self.clear_buffer()

    @property
    def closed(self) -> bool:
        return False

    def fileno(self) -> int:
        return hash(self)

    def isatty(self) -> bool:
        return False

    # ── Seek / tell ───────────────────────────────────────────────────

    def tell(self) -> int:
        return self.position

    def seekable(self) -> bool:
        return True

    def seek(self, offset, whence=0, /):
        if whence == io.SEEK_SET:
            new = offset
        elif whence == io.SEEK_CUR:
            new = self.position + offset
        elif whence == io.SEEK_END:
            new = self.content_length + offset
        else:
            raise ValueError("Invalid whence value")
        if new < 0:
            raise ValueError("Negative seek position")
        if self._buffer is not None:
            self._buffer.seek(new, io.SEEK_SET)
        self.position = new
        return self.position

    # ── Read ──────────────────────────────────────────────────────────

    def readable(self) -> bool:
        return True

    @abstractmethod
    def read_byte_range(
        self, start: int, length: int, allow_not_found: bool = False
    ) -> bytes:
        """Read *length* bytes starting at *start* from the remote file."""

    def read_all_bytes(
        self, use_cache: bool = True, allow_not_found: bool = False
    ) -> bytes:
        """Return the full file contents, optionally caching in the buffer."""
        if use_cache and self._buffer is not None:
            cached = self._buffer.to_bytes()
            if len(cached) == self.content_length:
                return cached
            self._buffer.close()
            self._buffer = None

        data = self.read_byte_range(
            0, self.content_length, allow_not_found=allow_not_found,
        )
        self.content_length = len(data)

        if use_cache and self._buffer is None:
            self._buffer = _Buffer(data)
            self._buffer.seek(self.position, io.SEEK_SET)
        return data

    def getvalue(self) -> bytes:
        if self._buffer is not None:
            return self._buffer.to_bytes()
        return self.read_all_bytes()

    def getbuffer(self) -> _Buffer:
        return self.buffer

    def read(self, n: int = -1, use_cache: bool = True):
        cur = self.position
        blob = self.read_all_bytes(use_cache=use_cache)
        if n == -1:
            n = self.content_length - cur
        chunk = blob[cur:cur + n]
        self.position += len(chunk)
        return chunk.decode(self.encoding) if self.encoding else chunk

    def readline(self, limit: int = -1, use_cache: bool = True):
        if self.encoding:
            chars = []
            read = 0
            while limit == -1 or read < limit:
                ch = self.read(1, use_cache=use_cache)
                if not ch:
                    break
                chars.append(ch)
                read += 1
                if ch == "\n":
                    break
            return "".join(chars)

        buf = bytearray()
        read = 0
        while limit == -1 or read < limit:
            b = self.read(1, use_cache=use_cache)
            if not b:
                break
            buf.extend(b)
            read += 1
            if b == b"\n":
                break
        return bytes(buf)

    def readlines(self, hint: int = -1, use_cache: bool = True):
        lines, total = [], 0
        while True:
            line = self.readline(use_cache=use_cache)
            if not line:
                break
            lines.append(line)
            total += len(line)
            if hint != -1 and total >= hint:
                break
        return lines

    # ── Write ─────────────────────────────────────────────────────────

    def writable(self) -> bool:
        return True

    @abstractmethod
    def write_all_bytes(self, data: Union[bytes, IO[bytes]]):
        """Write *data* to the remote path (replaces content)."""

    def truncate(self, size=None, /):
        if size is None:
            size = self.position
        if self._buffer is None:
            return self.write_all_bytes(data=b"\x00" * size)
        self._buffer.truncate(size)
        self.content_length = size
        self._write_flag = True
        return size

    def _need_flush(self) -> bool:
        return self._write_flag and self._buffer is not None

    def flush(self):
        if self._need_flush():
            self.write_all_bytes(data=self._buffer.to_bytes())
            self._write_flag = False

    def write(self, data: AnyStr) -> int:
        if isinstance(data, str):
            data = data.encode(self.encoding or "utf-8")
        written = self.buffer.write(data)
        self.position += written
        self.content_length = self.position
        self._write_flag = True
        return written

    def writelines(self, lines) -> None:
        for line in lines:
            if isinstance(line, str):
                line = line.encode(self.encoding or "utf-8")
            elif not isinstance(line, (bytes, bytearray)):
                raise TypeError(f"expected bytes-like or str, not {type(line).__name__!r}")
            payload = line if line.endswith(b"\n") else line + b"\n"
            self.write(payload)

    # ── Misc ──────────────────────────────────────────────────────────

    def size(self) -> int:
        return self.content_length

    def get_output_stream(self, *args, **kwargs):
        return self

    def copy_to(self, dest: Union["DatabricksIO", "DatabricksPath", str]) -> None:
        data = self.read_all_bytes(use_cache=False)
        if isinstance(dest, DatabricksIO):
            dest.write_all_bytes(data=data)
        elif hasattr(dest, "write"):
            dest.write(data)
        else:
            from .path import DatabricksPath
            dest_path = DatabricksPath.parse(dest, client=self.workspace)
            with dest_path.open(mode="wb") as d:
                self.copy_to(dest=d)

    def _reset_for_write(self):
        if self._buffer is not None:
            self._buffer.seek(0, io.SEEK_SET)
            self._buffer.truncate(0)
        else:
            self._buffer = _Buffer()
        self.position = 0
        self.content_length = 0
        self._write_flag = True


# ═══════════════════════════════════════════════════════════════════════════
# Workspace  (/Workspace/…)
# ═══════════════════════════════════════════════════════════════════════════

class DatabricksWorkspaceIO(DatabricksIO):
    """IO for Databricks Workspace files — Workspace API."""

    @retry(exceptions=(InternalError,))
    def read_byte_range(self, start: int, length: int, allow_not_found: bool = False) -> bytes:
        if length == 0:
            return b""
        sdk = self.workspace.workspace_client()
        result = sdk.workspace.download(
            path=self.path.full_path(),
            format=ExportFormat.AUTO,
        )
        if result is None:
            return b""
        data = result.read()
        return data[start:start + length]

    @retry(exceptions=(InternalError,))
    def write_all_bytes(self, data: Union[bytes, IO[bytes]]):
        sdk = self.workspace.workspace_client()
        full_path = self.path.full_path()

        if isinstance(data, bytes):
            bsize = len(data)
        elif isinstance(data, _Buffer):
            bsize = data.size
        elif isinstance(data, io.BytesIO):
            bsize = len(data.getvalue())
        else:
            bsize = None

        LOGGER.debug("Writing %s (size=%s) to %s", type(data).__name__, bsize, self)

        try:
            sdk.workspace.upload(full_path, data, format=ImportFormat.AUTO, overwrite=True)
        except (NotFound, ResourceDoesNotExist, BadRequest):
            self.path.parent.mkdir(parents=True, exist_ok=True)
            sdk.workspace.upload(full_path, data, format=ImportFormat.AUTO, overwrite=True)

        self.path.reset_metadata(is_file=True, is_dir=False, size=bsize, mtime=time.time())
        LOGGER.info("Written %s bytes to %s", bsize, self)
        return self


# ═══════════════════════════════════════════════════════════════════════════
# Volume  (/Volumes/…)
# ═══════════════════════════════════════════════════════════════════════════

class DatabricksVolumeIO(DatabricksIO):
    """IO for Unity Catalog volume files — Files API."""

    @retry(exceptions=(InternalError,))
    def read_byte_range(self, start: int, length: int, allow_not_found: bool = False) -> bytes:
        if length <= 0:
            return b""
        if start < 0:
            raise ValueError(f"start must be >= 0, got {start}")

        sdk = self.workspace.workspace_client()
        try:
            resp = sdk.files.download(self.path.full_path())
        except (NotFound, ResourceDoesNotExist, BadRequest, InternalError):
            if allow_not_found:
                return b""
            raise
        data = resp.contents.read()
        if start >= len(data):
            return b""
        return data[start:start + length]

    def write_all_bytes(
        self,
        data: Union[bytes, bytearray, memoryview, BinaryIO],
        *,
        overwrite: bool = True,
        part_size: Optional[int] = None,
        use_parallel: bool = True,
        parallelism: Optional[int] = None,
    ):
        sdk = self.workspace.workspace_client()
        full_path = self.path.full_path()

        LOGGER.debug("Writing all bytes to %s", self)
        size, bio, should_close = _prepare_buffer(data)

        def _upload():
            return sdk.files.upload(
                full_path, bio, overwrite=overwrite,
                part_size=part_size, use_parallel=use_parallel,
                parallelism=parallelism,
            )

        try:
            _upload()
        except (NotFound, ResourceDoesNotExist, BadRequest, InternalError):
            self.path.parent.mkdir(parents=True, exist_ok=True)
            try:
                bio.seek(0)
            except Exception:
                pass
            _upload()
        finally:
            if should_close:
                try:
                    bio.close()
                except Exception:
                    pass

        self.path.reset_metadata(is_file=True, is_dir=False, size=size, mtime=time.time())
        LOGGER.info("Written %s bytes to %s", size or "all", self.path)
        return self


# ═══════════════════════════════════════════════════════════════════════════
# DBFS  (/dbfs/…)
# ═══════════════════════════════════════════════════════════════════════════

class DatabricksDBFSIO(DatabricksIO):
    """IO for DBFS files — DBFS API with base-64 chunked reads."""

    @retry(exceptions=(InternalError,))
    def read_byte_range(self, start: int, length: int, allow_not_found: bool = False) -> bytes:
        if length == 0:
            return b""
        sdk = self.workspace.workspace_client()
        full_path = self.path.full_path()

        result = bytearray()
        remaining = length
        pos = start
        while remaining > 0:
            chunk_size = min(remaining, 2 * 1024 * 1024)
            resp = sdk.dbfs.read(path=full_path, offset=pos, length=chunk_size)
            if not resp.data:
                break
            decoded = base64.b64decode(resp.data)
            result.extend(decoded)
            pos += len(decoded)
            remaining -= len(decoded)
        return bytes(result)

    @retry(exceptions=(InternalError,))
    def write_all_bytes(self, data: Union[bytes, IO[bytes]]):
        sdk = self.workspace.workspace_client()
        full_path = self.path.full_path()

        LOGGER.debug("Writing all bytes to %s", self)
        try:
            with sdk.dbfs.open(path=full_path, read=False, write=True, overwrite=True) as f:
                f.write(data)
        except (NotFound, ResourceDoesNotExist, BadRequest):
            self.path.parent.mkdir(parents=True, exist_ok=True)
            with sdk.dbfs.open(path=full_path, read=False, write=True, overwrite=True) as f:
                f.write(data)

        LOGGER.info("Written all bytes to %s", self)
        self.path.reset_metadata(is_file=True, is_dir=False, size=len(data), mtime=time.time())




"""
Collection-level service for Unity Catalog tables.

Provides ``find_table`` and ``list_tables`` against the Databricks catalog API.
Per-table DDL/DML lives in :class:`~yggdrasil.databricks.sql.table.Table`.

Caching strategy
----------------
A module-level :class:`ExpiringDict` (``_TABLE_INFO_CACHE``) keyed by
``"host|catalog.schema.table"`` acts as a fast *local* cache so the same
table lookup never hits the API twice within the TTL window.

    1. **Local** — check ``_TABLE_INFO_CACHE``; return immediately on hit.
    2. **Remote** — call ``_find_table_remote``; only reached on miss.
    3. **Update** — store the remote result back into the local cache.

Cache entries can be invalidated per-table via :meth:`Tables._invalidate`.
"""

from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Optional, Iterator

from databricks.sdk.errors import DatabricksError, ResourceDoesNotExist
from databricks.sdk.service.catalog import TableInfo
from yggdrasil.databricks.client import DatabricksService
from yggdrasil.databricks.sql.types import quote_ident
from yggdrasil.dataclasses.expiring import ExpiringDict

from .table import Table

__all__ = ["Tables"]

logger = logging.getLogger(__name__)

# ---------------------------------------------------------------------------
# Module-level function cache
# Keyed by "host|catalog.schema.table"; default TTL = 5 minutes.
# ---------------------------------------------------------------------------
_TABLE_INFO_CACHE: ExpiringDict[str, TableInfo] = ExpiringDict(default_ttl=300.0)


@dataclass(frozen=True)
class Tables(DatabricksService):
    """Collection-level service for Unity Catalog tables.

    Attach default catalog / schema context so callers don't have to repeat
    them on every call::

        tables = client.tables(catalog_name="main", schema_name="sales")
        table  = tables.find_table("orders")
        for t in tables.list_tables():
            ...
    """

    catalog_name: Optional[str] = None
    schema_name: Optional[str] = None
    table_name: Optional[str] = None

    def parse_catalog_schema_table_names(self, full_name: str) -> tuple[Optional[str], Optional[str], Optional[str]]:
        parts = [_.strip("`") for _ in full_name.split(".")]

        # Unreachable for .split("."), but harmless and keeps intent explicit.
        if len(parts) == 0:
            return self.catalog_name, self.schema_name, None
        if len(parts) == 1:
            return self.catalog_name, self.schema_name, parts[0]
        if len(parts) == 2:
            return self.catalog_name, parts[0], parts[1]

        catalog_name, schema_name, table_name = parts[-3], parts[-2], parts[-1]
        return catalog_name or self.catalog_name, schema_name or self.schema_name, table_name

    def parse_check_location_params(
        self,
        location: Optional[str] = None,
        catalog_name: Optional[str] = None,
        schema_name: Optional[str] = None,
        table_name: Optional[str] = None,
        safe_chars: bool = True,
    ) -> tuple[str, Optional[str], Optional[str], Optional[str]]:
        if location:
            c, s, t = self.parse_catalog_schema_table_names(location)
            catalog_name, schema_name, table_name = catalog_name or c, schema_name or s, table_name or t

        catalog_name = catalog_name or self.catalog_name
        schema_name = schema_name or self.schema_name

        assert catalog_name, "No catalog name given"
        assert schema_name, "No schema name given"
        assert table_name, "No table name given"

        if safe_chars:
            location = f"{quote_ident(catalog_name)}.{quote_ident(schema_name)}.{quote_ident(table_name)}"
        else:
            location = f"{catalog_name}.{schema_name}.{table_name}"

        return location, catalog_name or self.catalog_name, schema_name or self.schema_name, table_name

    # -------------------------------------------------------------------------
    # Factory — create a Table resource bound to this service
    # -------------------------------------------------------------------------

    def table(
        self,
        location: Optional[str] = None,
        *,
        table_name: Optional[str] = None,
        catalog_name: Optional[str] = None,
        schema_name: Optional[str] = None,
    ) -> "Table":
        """Return a :class:`~yggdrasil.databricks.sql.table.Table` bound to this service."""

        return Table.parse_str(
            location=location,
            service=self,
            catalog_name=catalog_name or self.catalog_name,
            schema_name=schema_name or self.schema_name,
            table_name=table_name or self.table_name,
        )

    # -------------------------------------------------------------------------
    # Cache helpers
    # -------------------------------------------------------------------------

    def _cache_key(
        self,
        catalog_name: Optional[str],
        schema_name: Optional[str],
        table_name: Optional[str],
    ) -> str:
        """Build a stable, host-scoped cache key."""
        host = self.client.base_url.to_string() if self.client else "default"
        return f"{host}|{catalog_name}.{schema_name}.{table_name}"

    def _invalidate(
        self,
        catalog_name: Optional[str],
        schema_name: Optional[str],
        table_name: Optional[str],
    ) -> None:
        """Evict one entry from the module-level cache."""
        key = self._cache_key(catalog_name, schema_name, table_name)
        try:
            del _TABLE_INFO_CACHE[key]
        except KeyError:
            pass

    @classmethod
    def invalidate_all(cls) -> None:
        """Clear the entire module-level table-info cache."""
        _TABLE_INFO_CACHE.clear()

    # -------------------------------------------------------------------------
    # Remote fetch — single responsibility, no caching
    # -------------------------------------------------------------------------

    def find_table_remote(
        self,
        catalog_name: str,
        schema_name: str,
        table_name: Optional[str] = None,
        *,
        table_id: Optional[str] = None,
        raise_error: bool = True,
    ) -> Optional[TableInfo]:
        """Raw API lookup — three strategies in order, no cache.

        1. Search by ``table_id`` (full list scan, preferred when id is known).
        2. GET by fully-qualified name (fast path for normal lookups).
        3. Case-insensitive list scan (handles edge cases in naming / quoting).

        Returns ``None`` on miss when ``raise_error=False``.
        """
        uc = self.client.workspace_client().tables

        # --- Strategy 1: search by table_id ---------------------------------
        if table_id:
            logger.debug(
                "Remote fetch [strategy=table_id] id=%s catalog=%s schema=%s",
                table_id, catalog_name, schema_name,
            )
            try:
                for info in uc.list(catalog_name=catalog_name, schema_name=schema_name):
                    if info.table_id == table_id:
                        return info
            except DatabricksError as exc:
                if raise_error:
                    raise ResourceDoesNotExist(
                        f"Failed searching table_id={table_id} in"
                        f" {catalog_name}.{schema_name}"
                    ) from exc
                return None

            if raise_error:
                raise ResourceDoesNotExist(
                    f"Table id={table_id} not found in {catalog_name}.{schema_name}"
                )
            return None

        # --- Strategy 2: GET by full name (fast path) -----------------------
        full_name = (
            table_name
            if isinstance(table_name, str) and table_name.count(".") == 2
            else f"{catalog_name}.{schema_name}.{table_name}"
        )
        logger.debug("Remote fetch [strategy=get] full_name=%s", full_name)
        try:
            return uc.get(full_name=full_name)
        except DatabricksError:
            pass  # fall through to list scan
        except Exception as exc:
            if raise_error:
                raise ResourceDoesNotExist(f"Failed to GET table {full_name}") from exc
            return None

        # --- Strategy 3: case-insensitive list scan -------------------------
        logger.debug(
            "Remote fetch [strategy=list_scan] catalog=%s schema=%s name=%s",
            catalog_name, schema_name, table_name,
        )
        try:
            for info in uc.list(catalog_name=catalog_name, schema_name=schema_name):
                if info.name == table_name or info.name.lower() == str(table_name).lower():
                    return info
        except DatabricksError as exc:
            if raise_error:
                raise ResourceDoesNotExist(
                    f"Failed to list tables in {catalog_name}.{schema_name}"
                ) from exc
            return None

        if raise_error:
            raise ResourceDoesNotExist(f"Table {full_name} not found")
        return None

    # -------------------------------------------------------------------------
    # Public API
    # -------------------------------------------------------------------------

    def find_table(
        self,
        location: Optional[str] = None,
        *,
        catalog_name: Optional[str] = None,
        schema_name: Optional[str] = None,
        table_name: Optional[str] = None,
        table_id: Optional[str] = None,
        raise_error: bool = True,
        use_cache: bool = True,
        cache_ttl: Optional[float] = None,
    ) -> Optional["Table"]:
        """Resolve a table by name or Unity Catalog ID.

        Caching is enabled by default (``use_cache=True``).  Set
        ``use_cache=False`` to force a fresh API call and refresh the entry.

        Args:
            location:     Full string location
            table_name:   Name.
            catalog_name: Override catalog (falls back to service default).
            schema_name:  Override schema (falls back to service default).
            table_id:     Unity Catalog table UUID — triggers id-based search.
            raise_error:  Raise :exc:`ResourceDoesNotExist` when not found.
            use_cache:    ``False`` bypasses and refreshes the cache entry.
            cache_ttl:    Entry TTL in seconds (``None`` → 5 min default).
        """
        _, catalog, schema, name = self.parse_check_location_params(
            location=location, catalog_name=catalog_name, schema_name=schema_name,
            table_name=table_name
        )
        cache_key = self._cache_key(catalog, schema, name)

        # 1. Check local cache -----------------------------------------------
        if use_cache:
            cached: Optional[TableInfo] = _TABLE_INFO_CACHE.get(cache_key)
            if cached is not None:
                logger.debug(
                    "Cache hit [Tables.find_table] key=%s table_id=%s",
                    cache_key, cached.table_id,
                )
                tb = Table(
                    service=self,
                    catalog_name=cached.catalog_name,
                    schema_name=cached.schema_name,
                    table_name=cached.name,
                )
                object.__setattr__(tb, "_infos", cached)
                return tb

        # 2. Fetch remote ----------------------------------------------------
        info = self.find_table_remote(
            catalog_name=catalog,
            schema_name=schema,
            table_name=name,
            table_id=table_id,
            raise_error=raise_error,
        )
        if info is None:
            return None

        # 3. Update local cache ----------------------------------------------
        if use_cache:
            if cache_ttl is not None:
                _TABLE_INFO_CACHE.set(cache_key, info, ttl=cache_ttl)
            else:
                _TABLE_INFO_CACHE.set(cache_key, info)

        tb = Table(
            service=self,
            catalog_name=info.catalog_name,
            schema_name=info.schema_name,
            table_name=info.name,
        )
        object.__setattr__(tb, "_infos", info)
        return tb

    def list_tables(
        self,
        catalog_name: Optional[str] = None,
        schema_name: Optional[str] = None,
        *,
        use_cache: bool = True,
        cache_ttl: Optional[float] = None,
    ) -> Iterator["Table"]:
        """Iterate over all tables in a schema, populating the local cache.

        Args:
            catalog_name: Override catalog (falls back to service default).
            schema_name:  Override schema (falls back to service default).
            use_cache:    Populate ``_TABLE_INFO_CACHE`` from results.
            cache_ttl:    Entry TTL in seconds (``None`` → 5 min default).
        """

        catalog_name = catalog_name or self.catalog_name
        schema_name = schema_name or self.schema_name
        uc = self.client.workspace_client().tables

        for info in uc.list(catalog_name=catalog_name, schema_name=schema_name):
            if use_cache:
                key = self._cache_key(info.catalog_name, info.schema_name, info.name)
                if _TABLE_INFO_CACHE.get(key) is None:
                    if cache_ttl is not None:
                        _TABLE_INFO_CACHE.set(key, info, ttl=cache_ttl)
                    else:
                        _TABLE_INFO_CACHE.set(key, info)

            tb = Table(
                service=self,
                catalog_name=info.catalog_name,
                schema_name=info.schema_name,
                table_name=info.name,
            )
            object.__setattr__(tb, "_infos", info)
            yield tb




"""
Backward-compatible re-export façade for the complex serialization layer.

The implementation has been split into three focused modules:
    libs.py        — shared utilities and base serializer classes
    callables.py   — FunctionSerialized / MethodSerialized + helpers
    dataclasses.py — DataclassSerialized + helpers

Everything that was previously in this file is re-exported here so that all
existing import paths (``from yggdrasil.pickle.ser.complexs import …``) keep
working without modification.

``import inspect`` is kept at module level so that test code that does::

    import yggdrasil.pickle.ser.complexs as complexs_module
    monkeypatch.setattr(complexs_module.inspect, "getclosurevars", …)

continues to work — patching ``inspect.getclosurevars`` through this module
attribute affects the shared ``inspect`` module object used by callables.py.
"""

from __future__ import annotations

import inspect  # noqa: F401 — kept for test monkeypatching; see module docstring
from types import FunctionType, MethodType, ModuleType

from yggdrasil.pickle.ser.callables import (
    FunctionSerialized,
    MethodSerialized,
)
from yggdrasil.pickle.ser.dataclasses import (
    DataclassSerialized,
)
from yggdrasil.pickle.ser.libs import (
    BaseExceptionSerialized,
    ClassSerialized,
    ComplexSerialized,
    ModuleSerialized,
)

# ---------------------------------------------------------------------------
# public API
# ---------------------------------------------------------------------------

__all__ = [
    "ComplexSerialized",
    "ModuleSerialized",
    "ClassSerialized",
    "FunctionSerialized",
    "MethodSerialized",
    "BaseExceptionSerialized",
    "DataclassSerialized",
]

# ---------------------------------------------------------------------------
# Tags registration
# ---------------------------------------------------------------------------

from yggdrasil.pickle.ser.tags import Tags  # noqa: E402 — must come after class definitions

for _cls in ComplexSerialized.__subclasses__():
    Tags.register_class(_cls)

for _pytype, _cls in (
    (ModuleType, ModuleSerialized),
    (FunctionType, FunctionSerialized),
    (MethodType, MethodSerialized),
    (BaseException, BaseExceptionSerialized),
):
    Tags.register_class(_cls, pytype=_pytype)

del _cls, _pytype




"""
Job — immutable unit of work.

Thread-specific classes (AsyncJob, ThreadJob) live in
:mod:`yggdrasil.concurrent.threading`.
"""

from __future__ import annotations

import logging
from dataclasses import dataclass, field
from typing import Any, Callable, Dict, Generic, Optional, Tuple, TypeVar, TYPE_CHECKING

if TYPE_CHECKING:
    from yggdrasil.concurrent.threading import ThreadJob

__all__ = ["Job"]

LOGGER = logging.getLogger(__name__)

T = TypeVar("T")


@dataclass(frozen=True, slots=True)
class Job(Generic[T]):
    """Immutable, callable bundle describing a unit of work.

    Create with :meth:`make` rather than the constructor directly::

        job = Job.make(my_func, arg1, arg2, key=val)
        result = job.run()                # synchronous
        handle = job.fire_and_forget()    # background thread, returns ThreadJob
        handle.wait()                     # block until done
    """

    func: Callable[..., T]
    args: Tuple[Any, ...] = field(default_factory=tuple)
    kwargs: Dict[str, Any] = field(default_factory=dict)

    def __call__(self, *args: Any, **kwargs: Any) -> T:
        merged_args = self.args + args
        merged_kwargs = {**self.kwargs, **kwargs}
        return self.func(*merged_args, **merged_kwargs)

    @classmethod
    def make(
        cls,
        func: Callable[..., T],
        *args: Any,
        **kwargs: Any,
    ) -> "Job[T]":
        """Build a :class:`Job` from a callable and its arguments."""
        return cls(func=func, args=args, kwargs=kwargs)

    def run(self) -> T:
        """Execute the job synchronously in the calling thread."""
        return self.func(*self.args, **self.kwargs)

    def thread(self) -> "ThreadJob[T]":
        """Start this job in a daemon thread and return an awaitable :class:`~yggdrasil.concurrent.threading.ThreadJob`.

        The thread is started immediately.  Call :meth:`~yggdrasil.concurrent.threading.ThreadJob.wait`
        to block until completion and retrieve the result.
        """
        # Lazy import avoids the job → threading → job circular dependency.
        from yggdrasil.concurrent.threading import ThreadJob  # noqa: PLC0415
        return ThreadJob(job=self)

    def fire_and_forget(self) -> "ThreadJob[T]":
        """Alias for :meth:`thread`.

        Returns the :class:`~yggdrasil.concurrent.threading.ThreadJob` handle so callers
        can optionally await the result — or simply ignore it for true fire-and-forget.
        """
        return self.thread()



"""
Bounded thread-pool for large / infinite job streams.
"""

from __future__ import annotations

import logging
import os
from collections import deque
from concurrent.futures import ThreadPoolExecutor, Future, wait, FIRST_COMPLETED
from typing import Any, Deque, Iterable, Iterator, Optional, Set

from .job import Job
from .job_result import JobResult

__all__ = ["JobPoolExecutor"]

LOGGER = logging.getLogger(__name__)


class JobPoolExecutor(ThreadPoolExecutor):
    """ThreadPoolExecutor helper for large / infinite :class:`Job` streams.

    Keeps at most *max_in_flight* futures submitted at any time so the caller
    is never overwhelmed when the job source is unbounded.

    Args:
        max_workers:     Thread-pool size (``None`` → ``os.cpu_count()``).
        max_in_flight:   Max in-flight futures (``None`` → ``max_workers * 2``).
        job_name_prefix: Thread-name prefix forwarded to ``ThreadPoolExecutor``.

    Yields (via :meth:`as_completed`):
        :class:`JobResult` in completion order (``ordered=False``) or
        submission order (``ordered=True``).
    """

    def __init__(
        self,
        max_workers: Optional[int] = None,
        max_in_flight: Optional[int] = None,
        job_name_prefix: str = "",
    ) -> None:
        super().__init__(
            max_workers=max_workers,
            thread_name_prefix=job_name_prefix,
        )
        self.job_name_prefix = job_name_prefix
        self.max_in_flight: int = max_in_flight if max_in_flight else (self._max_workers * 2)

    # ------------------------------------------------------------------
    # Submission helpers
    # ------------------------------------------------------------------

    def submit_job(self, job: Job) -> Future:
        """Submit a single :class:`Job` and return its :class:`Future`."""
        return self.submit(job.run)

    def _try_submit_next(self, it: Iterator[Job]) -> Optional[Future]:
        try:
            job = next(it)
        except StopIteration:
            return None
        return self.submit_job(job)

    # ------------------------------------------------------------------
    # Constructors
    # ------------------------------------------------------------------

    @classmethod
    def parse(
        cls,
        obj: Any,
        *,
        max_workers: Optional[int] = None,
    ) -> "JobPoolExecutor":
        if isinstance(obj, cls):
            return cls  # type: ignore[return-value]
        return cls(max_workers=max_workers or os.cpu_count())

    # ------------------------------------------------------------------
    # Properties
    # ------------------------------------------------------------------

    @property
    def max_workers(self) -> int:
        return self._max_workers

    # ------------------------------------------------------------------
    # Internal helpers
    # ------------------------------------------------------------------

    @staticmethod
    def _cancel_all(fs: Iterable[Future]) -> None:
        """Best-effort cancellation of pending (not-yet-started) futures."""
        for f in fs:
            if not f.done():
                f.cancel()

    def _unpack(
        self,
        fut: Future,
        raise_error: bool,
        pending: Iterable[Future],
    ) -> JobResult:
        exc = fut.exception()
        if exc is not None:
            if raise_error:
                self._cancel_all(pending)
                raise exc
            LOGGER.exception(exc)
            return JobResult(None, exc)
        return JobResult(fut.result(), None)

    # ------------------------------------------------------------------
    # Public streaming API
    # ------------------------------------------------------------------

    def as_completed(
        self,
        jobs: Iterable[Job],
        *,
        ordered: bool = False,
        max_in_flight: Optional[int] = None,
        cancel_on_exit: bool = False,
        shutdown_on_exit: bool = False,
        shutdown_wait: bool = False,
        raise_error: bool = True,
    ) -> Iterator[JobResult]:
        """Consume a (possibly infinite) :class:`Job` iterable and yield :class:`JobResult` objects.

        Args:
            jobs:            Source of :class:`Job` objects (may be infinite).
            ordered:         ``False`` (default) → completion order.
                             ``True``            → strict submission order.
            max_in_flight:   Override the instance-level window for this call.
            cancel_on_exit:  Cancel pending futures on generator close / error.
            shutdown_on_exit: Shut down the pool on generator close / error.
            shutdown_wait:   Wait for running threads when shutting down.
            raise_error:     Re-raise job exceptions at the yield site.

        Yields:
            :class:`JobResult` — one per submitted job.
        """
        window = max_in_flight or self.max_in_flight
        it = iter(jobs)

        if ordered:
            inflight: Deque[Future] = deque()
            try:
                for _ in range(window):
                    fut = self._try_submit_next(it)
                    if fut is None:
                        break
                    inflight.append(fut)

                while inflight:
                    head = inflight[0]
                    if not head.done():
                        wait({head}, return_when=FIRST_COMPLETED)

                    job_result = self._unpack(
                        inflight.popleft(),
                        raise_error=raise_error,
                        pending=inflight,
                    )
                    if job_result.ok:
                        yield job_result
                    elif raise_error:
                        raise job_result.exception  # type: ignore[misc]

                    nxt = self._try_submit_next(it)
                    if nxt is not None:
                        inflight.append(nxt)

            finally:
                if cancel_on_exit:
                    self._cancel_all(inflight)
                if shutdown_on_exit:
                    self.shutdown(wait=shutdown_wait, cancel_futures=True)

        else:
            pending: Set[Future] = set()
            try:
                for _ in range(window):
                    fut = self._try_submit_next(it)
                    if fut is None:
                        break
                    pending.add(fut)

                while pending:
                    done, _ = wait(pending, return_when=FIRST_COMPLETED)
                    for fut in done:
                        pending.discard(fut)

                        job_result = self._unpack(
                            fut,
                            raise_error=raise_error,
                            pending=pending,
                        )
                        if job_result.ok:
                            yield job_result
                        elif raise_error:
                            raise job_result.exception  # type: ignore[misc]

                        nxt = self._try_submit_next(it)
                        if nxt is not None:
                            pending.add(nxt)

            finally:
                if cancel_on_exit:
                    self._cancel_all(pending)
                if shutdown_on_exit:
                    self.shutdown(wait=shutdown_wait, cancel_futures=True)




from __future__ import annotations

from .cities import load_cities
from .core import load_core_geozones
from .country_zones import load_country_zones
from .countries import load_countries
from .geozone import GeoZone
from .mountains import load_mountains
from .zones_entsoe import load_entsoe_zones
from .zones_markets import load_market_zones

__all__ = ['load_geozones']

_LOADERS = (
    load_core_geozones,
    load_countries,
    load_cities,
    load_entsoe_zones,
    load_market_zones,
    load_country_zones,
    load_mountains,
)


def load_geozones() -> None:
    GeoZone.clear_cache()
    for loader in _LOADERS:
        loader()
    GeoZone.UK = GeoZone.UNITED_KINGDOM
    GeoZone.USA = GeoZone.UNITED_STATES
    GeoZone.US = GeoZone.UNITED_STATES



"""Concrete HTTP/HTTPS session backed by ``urllib3``."""
from __future__ import annotations

import datetime as dt
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Optional

import urllib3

from yggdrasil.concurrent.threading import Job, JobPoolExecutor
from yggdrasil.dataclasses import WaitingConfig
from yggdrasil.io import BytesIO, MediaType, MimeTypes
from .response import HTTPResponse
from ..request import PreparedRequest
from ..send_config import SendConfig
from ..session import Session

if TYPE_CHECKING:
    from .browser import BrowserHTTPSession

__all__ = ["HTTPSession"]


@dataclass
class HTTPSession(Session):
    """Concrete HTTP/HTTPS session backed by a ``urllib3`` connection pool."""

    _http_pool: urllib3.PoolManager = field(default=None, init=False, repr=False, compare=False)

    def _build_http_pool(self) -> urllib3.PoolManager:
        retries = urllib3.Retry(
            total=4,
            connect=2,
            read=2,
            backoff_factor=2,
            status_forcelist=(429, 500, 502, 503, 504),
            raise_on_status=False,
        )

        return urllib3.PoolManager(
            num_pools=self.pool_maxsize,
            maxsize=self.pool_maxsize,
            block=True,
            retries=retries,
            cert_reqs="CERT_REQUIRED" if self.verify else "CERT_NONE",
            ca_certs=None,
        )

    def __post_init__(self):
        if self.pool_maxsize:
            self.pool_maxsize = min(8, int(self.pool_maxsize))
        else:
            self.pool_maxsize = 8

        super().__post_init__()

        if self._http_pool is None:
            self._http_pool = self._build_http_pool()

    @property
    def http_pool(self):
        if self._http_pool is None:
            with self._lock:
                if self._http_pool is None:
                    self._http_pool = self._build_http_pool()
        return self._http_pool

    # ------------------------------------------------------------------
    # Extensibility hooks
    # ------------------------------------------------------------------

    def _build_request_headers(
        self,
        request: PreparedRequest,
    ) -> Optional[dict[str, str]]:
        """Return the headers dict to merge into *request* before sending.

        Subclasses may override this to inject per-request headers without
        replacing the entire :attr:`send_headers` mapping.  The default
        implementation returns :attr:`send_headers` unchanged.
        """
        return self.send_headers

    def to_browser(
        self,
        *,
        user_agent: Optional[str] = None,
        accept_language: str = "en-US,en;q=0.9",
        ua_seed: Optional[int] = None,
    ) -> "BrowserHTTPSession":
        """Return a :class:`~yggdrasil.io.http_.browser.BrowserHTTPSession`
        that inherits this session's connection settings.

        Parameters
        ----------
        user_agent:
            Explicit ``User-Agent`` string.  ``None`` (default) = auto-generate.
        accept_language:
            ``Accept-Language`` header value for browser requests.
        ua_seed:
            Optional integer seed for deterministic user-agent generation.

        Returns
        -------
        BrowserHTTPSession
        """
        from .browser import BrowserHTTPSession

        return BrowserHTTPSession(
            base_url=self.base_url,
            verify=self.verify,
            pool_maxsize=self.pool_maxsize,
            send_headers=dict(self.send_headers) if self.send_headers else None,
            waiting=self.waiting,
            user_agent=user_agent,
            accept_language=accept_language,
            ua_seed=ua_seed,
        )

    def _local_send(
        self,
        request: PreparedRequest,
        config: SendConfig,
    ) -> HTTPResponse:
        wait_cfg = self.waiting if config.wait is None else config.wait

        request = request.prepare_to_send(
            sent_at=None,
            headers=self._build_request_headers(request),
        )

        raw_resp = self.http_pool.request(
            method=request.method,
            url=request.url.to_string(),
            body=request.buffer,
            headers=request.headers,
            timeout=wait_cfg.timeout_urllib3,
            preload_content=False,
            decode_content=False,
            redirect=True,
        )

        result = HTTPResponse.from_urllib3(
            request=request,
            response=raw_resp,
            tags=None,
            received_at=dt.datetime.now(dt.timezone.utc),
        )
        result.drain_urllib3(raw_resp, stream=True, release_conn=True)

        x_current_page = raw_resp.headers.get("X-Current-Page")
        x_total_pages = raw_resp.headers.get("X-Last-Page")

        if x_current_page and x_total_pages:
            result = self._combine_paginated_pages(
                result=result,
                request=request,
                current_page=int(x_current_page),
                total_pages=int(x_total_pages),
                wait_cfg=wait_cfg,
                stream=config.stream,
                raise_error=config.raise_error,
            )

        if config.raise_error:
            result.raise_for_status()

        return result

    def _fetch_paginated_page(
        self,
        *,
        request: PreparedRequest,
        page_num: int,
        body_seed: bytes | None,
        wait_cfg: WaitingConfig,
        stream: bool,
        raise_error: bool,
    ) -> tuple[int, HTTPResponse]:
        page_url = request.url.add_query_item("page", str(page_num), replace=True)

        page_request = request.copy(
            url=page_url,
            buffer=BytesIO(body_seed) if body_seed is not None else None,
        )

        raw_resp = self.http_pool.request(
            method=page_request.method,
            url=page_url.to_string(),
            body=page_request.buffer,
            headers=page_request.headers,
            timeout=wait_cfg.timeout_urllib3,
            preload_content=not stream,
            decode_content=False,
            redirect=True,
        )

        page_result = HTTPResponse.from_urllib3(
            request=page_request,
            response=raw_resp,
            tags=None,
            received_at=dt.datetime.now(tz=dt.timezone.utc),
        )
        page_result.drain_urllib3(raw_resp, stream=stream, release_conn=True)

        if raise_error:
            page_result.raise_for_status()

        return page_num, page_result

    def _combine_paginated_pages(
        self,
        *,
        result: HTTPResponse,
        request: PreparedRequest,
        current_page: int,
        total_pages: int,
        wait_cfg: WaitingConfig,
        stream: bool,
        raise_error: bool,
        pool: Optional[JobPoolExecutor | int] = None,
    ) -> HTTPResponse:
        if not isinstance(pool, JobPoolExecutor):
            with JobPoolExecutor.parse(pool) as parsed_pool:
                return self._combine_paginated_pages(
                    result=result,
                    request=request,
                    current_page=current_page,
                    total_pages=total_pages,
                    wait_cfg=wait_cfg,
                    stream=stream,
                    raise_error=raise_error,
                    pool=parsed_pool,
                )

        from yggdrasil.polars.lib import polars as pl

        init_df = result.to_polars(parse=True, lazy=False)
        if total_pages <= current_page:
            return result

        remaining_pages = list(range(current_page + 1, total_pages + 1))
        body_seed = request.buffer.to_bytes() if request.buffer else None

        def jobs():
            for pn in remaining_pages:
                yield Job.make(
                    self._fetch_paginated_page,
                    request=request,
                    page_num=pn,
                    body_seed=body_seed,
                    wait_cfg=wait_cfg,
                    stream=stream,
                    raise_error=raise_error,
                )

        frames = [init_df]
        for job_result in pool.as_completed(
            jobs(),
            ordered=False,
            max_in_flight=len(remaining_pages),
            cancel_on_exit=False,
            shutdown_on_exit=False,
            raise_error=True,
        ):
            _, page_resp = job_result.result
            frames.append(page_resp.to_polars(parse=True, lazy=False))

        final_df = pl.concat(frames, how="diagonal_relaxed", rechunk=True)

        result.buffer.truncate(size=0)

        arr = final_df.to_arrow(compat_level=pl.CompatLevel.newest())
        mt = MediaType(MimeTypes.ARROW_IPC)
        mio = result.buffer.media_io(mt)
        mio.write_arrow_table(arr)
        result.set_media_type(mt, safe=False)

        result.update_tags({
            "page_start": str(current_page),
            "page_total": str(total_pages),
        })

        return result